
from __future__ import annotations
from typing import Any, Mapping

from ..models.user_relationship_item import UserRelationshipItem, FriendRelationshipItem, BannedUserRelationshipItem

def load_user_relationship_item(d: Mapping[str, Any]) -> UserRelationshipItem:
    return UserRelationshipItem(d)

def load_friend_relationship_item(d: Mapping[str, Any]) -> FriendRelationshipItem:
    return FriendRelationshipItem(d)

def load_banned_user_relation_item(d: Mapping[str, Any]) -> BannedUserRelationshipItem:
    return BannedUserRelationshipItem(d)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Sequence, Mapping
if TYPE_CHECKING:
    from .http.response import Response

from . import http

class ArgExcMixin(Exception):
    def __init__(self, arg: object = None) -> None:
        super().__init__()
        self.arg: object = arg

    def __str__(self) -> str:
        if self.arg is None:
            return self.get_default_message()
        return str(self.arg)

    def get_default_message(self) -> str:
        return ''

class ArgExc(ArgExcMixin):
    pass


class ClientException(ArgExc):
    pass

class NoResultException(ClientException):
    pass

class RejectedResultException(ClientException):
    pass

class UnexpectedResultException(ClientException):
    pass


class _Throwaway(ArgExc):
    pass

class UserAgentRequired(_Throwaway):
    pass


def raise_for_non_json_response(resp: Response) -> None:
    data = resp.data
    is_html_content = resp.headers.get('Content-Type', '').startswith('text/html')

    if is_html_content:
        if b"user agent required" in data:
            raise UserAgentRequired('the Reddit API wants you to set a user agent')

    try:
        resp.raise_for_status()
    except http.exceptions.StatusCodeException as e:
        if is_html_content:
            msg = None
            if b"Our CDN was unable to reach our servers" in data:
                msg = 'HTML, "Our CDN was unable to reach our servers"'
            if b"title>reddit.com: page not found</title" in data:
                msg = 'HTML, page not found'
            e.arg = msg
        raise


class APIError(ArgExc):
    pass

class RedditError(APIError):
    def __init__(self,
        arg: object = None,
        *,
        codename: str,
        explanation: str,
        field: str,
    ) -> None:
        super().__init__(arg)
        self.codename: str = codename
        self.explanation: str = explanation
        self.field: str = field

    def get_default_message(self) -> str:
        co = self.codename
        xp = self.explanation
        fd = self.field
        if co:
            if xp:
                if fd:
                    return f'{co}: {xp} -> {fd}'
                return f'{co}: {xp}'
            if fd:
                return f'{co} -> {fd}'
            return co
        return ''


def raise_for_reddit_error(json_data: Any) -> None:
    if not isinstance(json_data, Mapping):
        return

    error_record: Sequence[Any]
    if (
        isinstance(codename := json_data.get('reason'), str)
        and isinstance(explanation := json_data.get('explanation'), str)
        and isinstance(field := next(iter(json_data.get('fields', [])), None), str)
    ):
        raise RedditError(codename=codename, explanation=explanation, field=field)
    elif (
        isinstance(codename := json_data.get('reason'), str)
        and json_data.get('explanation') is None
        and isinstance(field := next(iter(json_data.get('fields', [])), None), str)
    ):
        raise RedditError(codename=codename, explanation='', field=field)
    elif (
        isinstance(reason := json_data.get('reason'), str)
        and json_data.get('explanation') is None
        and list(json_data.get('fields', [])) == [None]
    ):
        raise RedditError(codename=reason, explanation='', field='')
    elif (
        isinstance(codename := json_data.get('reason'), str)
        and isinstance(explanation := json_data.get('explanation'), str)
    ):
        raise RedditError(codename=codename, explanation=explanation, field='')
    elif json_data.keys() >= {'error', 'message'} and isinstance(reason := json_data.get('reason'), str):
        raise RedditError(codename=reason, explanation='', field='')
    elif json_data.keys() >= {'error', 'message'}:
        return  # No useful information. Treat this as a StatusCodeException.
    elif json_data.keys() >= {'message'} and isinstance(reason := json_data.get('reason'), str):
        raise RedditError(codename=reason, explanation='', field='')
    elif (
        (error_record := next(iter(json_data.get('json', {}).get('errors', [])), [None, None, None]))
        and isinstance(codename := error_record[0], str)
        and isinstance(explanation := error_record[1], str)
        and isinstance(field := error_record[2], str)
    ):
        raise RedditError(codename=codename, explanation=explanation, field=field)
    elif (
        (error_record := next(iter(json_data.get('json', {}).get('errors', [])), [None, None, None]))
        and isinstance(codename := error_record[0], str)
        and isinstance(explanation := error_record[1], str)
    ):
        raise RedditError(codename=codename, explanation=explanation, field='')
    elif (
        isinstance(codename := next(iter(json_data.get('errors', [])), None), str)
        and isinstance(explanation := next(iter(json_data.get('errors_values', [])), None), str)
    ):
        raise RedditError(codename=codename, explanation=explanation, field='')




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_ASYNC import Subreddit

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.account_async1 import (
    SubscribedListingAsyncPaginator,
    ContributingListingAsyncPaginator,
    ModeratingListingAsyncPaginator,
)

class PullSubreddits:
    def __init__(self, client: Client):
        self._client = client

    def subscribed(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingAsyncIterator[SubscribedListingAsyncPaginator, Subreddit]:
        p = SubscribedListingAsyncPaginator(self._client, '/subreddits/mine/subscriber')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def contributing(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingAsyncIterator[ContributingListingAsyncPaginator, Subreddit]:
        p = ContributingListingAsyncPaginator(self._client, '/subreddits/mine/contributor')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def moderating(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingAsyncIterator[ModeratingListingAsyncPaginator, Subreddit]:
        p = ModeratingListingAsyncPaginator(self._client, '/subreddits/mine/moderator')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .....client_SYNC import Client
    from .....models.message_SYNC import MailboxMessage
    from ..stream import IStandardStreamEventSubject

from .....models.message_SYNC import ComposedMessage, CommentMessage
from ..stream import Stream


def make_unread_message_stream(client: Client) -> IStandardStreamEventSubject[MailboxMessage]:
    it = client.p.message.pull.unread()
    paginator = it.get_paginator()
    def extractor(message: MailboxMessage) -> tuple[int, int]:
        if isinstance(message, ComposedMessage):
            return (0, message.id)
        elif isinstance(message, CommentMessage):
            return (1, message.comment.id)
        raise Exception
    return Stream(paginator, extractor)

def make_mentions_message_stream(client: Client) -> IStandardStreamEventSubject[CommentMessage]:
    it = client.p.message.pull.mentions()
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.comment.id)




from __future__ import annotations
from typing import TYPE_CHECKING, IO, Optional, Literal, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_style_asset_upload_lease import SubredditStyleAssetUploadLease

from functools import cached_property

from ...model_loaders.subreddit_style_asset_upload_lease import load_subreddit_style_asset_upload_lease
from ...http.payload import guess_mimetype_from_filename

class RedesignRedditSubredditStyleProcedures:
    def __init__(self, client: Client):
        self._client = client

    class _upload_banner:
        def __init__(self, outer: RedesignRedditSubredditStyleProcedures) -> None:
            self._client = outer._client

        async def _obtain_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None, imagetype: str) -> SubredditStyleAssetUploadLease:
            if mimetype is None:
                mimetype = guess_mimetype_from_filename(filename)
            result = await self._client.request('POST', f'/api/v1/style_asset_upload_s3/{sr}',
                    data={'filepath': filename, 'mimetype': mimetype, 'imagetype': imagetype})
            return load_subreddit_style_asset_upload_lease(result)

        async def obtain_banner_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filename=filename, mimetype=mimetype, imagetype='bannerBackgroundImage')

        async def obtain_banner_overlay_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filename=filename, mimetype=mimetype, imagetype='bannerPositionedImage')

        async def obtain_banner_overlay_hover_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filename=filename, mimetype=mimetype, imagetype='secondaryBannerPositionedImage')

        async def obtain_mobile_banner_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filename=filename, mimetype=mimetype, imagetype='mobileBannerImage')

        async def deposit_file(self, file: IO[bytes], upload_lease: SubredditStyleAssetUploadLease, *,
                timeout: float = 1000) -> None:
            session = self._client.http.session
            resp = await session.request('POST', upload_lease.endpoint, data=upload_lease.fields,
                    files={'file': file}, timeout=timeout)
            resp.raise_for_status()

        async def upload_banner(self, file: IO[bytes], *, sr: str, timeout: float = 1000) -> SubredditStyleAssetUploadLease:
            upload_lease = await self.obtain_banner_upload_lease(filename=file.name, sr=sr)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def upload_banner_overlay(self, file: IO[bytes], *, sr: str, timeout: float = 1000) -> SubredditStyleAssetUploadLease:
            upload_lease = await self.obtain_banner_overlay_upload_lease(filename=file.name, sr=sr)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def upload_banner_overlay_hover(self, file: IO[bytes], *, sr: str, timeout: float = 1000) -> SubredditStyleAssetUploadLease:
            upload_lease = await self.obtain_banner_overlay_hover_upload_lease(filename=file.name, sr=sr)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def upload_mobile_banner(self, file: IO[bytes], *, sr: str, timeout: float = 1000) -> SubredditStyleAssetUploadLease:
            upload_lease = await self.obtain_mobile_banner_upload_lease(filename=file.name, sr=sr)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

    upload_banner: cached_property[_upload_banner] = cached_property(_upload_banner)

    async def configure_banner_settings(self,
        sr: str,
        *,
        banner_size: Optional[Literal['', 'small', 'medium', 'large']] = None,
        banner_background_color: Optional[str] = None,
        banner_image_url: Optional[str] = None,
        banner_image_display: Optional[Literal['', 'fill', 'tile']] = None,
        banner_overlay_image_url: Optional[str] = None,
        banner_overlay_image_position: Optional[Literal['', 'left', 'center', 'right']] = None,
        banner_overlay_hover_image_url: Optional[str] = None,
        mobile_banner_image_url: Optional[str] = None,
    ) -> None:
        def g() -> Iterable[tuple[str, str]]:
            if banner_size is not None:
                yield ('bannerHeight', banner_size)
            if banner_background_color is not None:
                yield ('bannerBackgroundColor', banner_background_color)
            if banner_image_url is not None:
                yield ('bannerBackgroundImage', banner_image_url)
            if banner_image_display is not None:
                v = {
                    '': '',
                    'fill': 'cover',
                    'tile': 'tiled',
                }[banner_image_display]
                yield ('bannerBackgroundImagePosition', v)
            if banner_overlay_image_url is not None:
                yield ('bannerPositionedImage', banner_overlay_image_url)
            if banner_overlay_image_position is not None:
                v = {
                    '': '',
                    'left': 'left',
                    'center': 'centered',
                    'right': 'right',
                }[banner_overlay_image_position]
                yield ('bannerPositionedImagePosition', v)
            if banner_overlay_hover_image_url is not None:
                yield ('secondaryBannerPositionedImage', banner_overlay_hover_image_url)
            if mobile_banner_image_url is not None:
                yield ('mobileBannerImage', mobile_banner_image_url)

        await self._client.request('PATCH', f'/api/v1/structured_styles/{sr}', data=dict(g()))




from __future__ import annotations

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class ModeratorInfo:
    name: str
    flair_type: str
    flair_text: str
    flair_fg_light_or_dark: str
    flair_bg_color: str
    flair_has_had_flair: bool




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_ASYNC import Submission

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.front_async1 import (
    BestListingAsyncPaginator,
    HotListingAsyncPaginator,
    NewListingAsyncPaginator,
    TopListingAsyncPaginator,
    RisingListingAsyncPaginator,
    ControversialListingAsyncPaginator,
    GildedListingAsyncPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def best(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[BestListingAsyncPaginator, Submission]:
        p = BestListingAsyncPaginator(self._client, '/best')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def hot(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[HotListingAsyncPaginator, Submission]:
        p = HotListingAsyncPaginator(self._client, '/hot')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def new(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[NewListingAsyncPaginator, Submission]:
        p = NewListingAsyncPaginator(self._client, '/new')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def top(self, amount: Optional[int] = None, *,
            time: str = 'day',
            ) -> ImpartedPaginatorChainingAsyncIterator[TopListingAsyncPaginator, Submission]:
        p = TopListingAsyncPaginator(self._client, '/top')
        p.time = time
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def rising(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[RisingListingAsyncPaginator, Submission]:
        p = RisingListingAsyncPaginator(self._client, '/rising')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def controversial(self, amount: Optional[int] = None, *,
            time: str = 'day',
        ) -> ImpartedPaginatorChainingAsyncIterator[ControversialListingAsyncPaginator, Submission]:
        p = ControversialListingAsyncPaginator(self._client, '/controversial')
        p.time = time
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def gilded(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[GildedListingAsyncPaginator, object]:
        p = GildedListingAsyncPaginator(self._client, '/gilded')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from typing import Any, Mapping, Optional

from datetime import datetime
from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class ModerationNote(IArtifact):
    #_: KW_ONLY
    d: Mapping[str, Any]
    uuid: str
    timestamp: int
    datetime: datetime
    type: str
    subreddit_id: int
    subreddit: str
    agent_id: int
    agent: str
    target_id: int
    target: str

@dataclass(repr=False, eq=False)
class ModerationActionNote(ModerationNote):
    action: str

@dataclass(repr=False, eq=False)
class ModerationUserNote(ModerationNote):
    note: str
    label: str
    has_anchor: bool
    anchor_submission_id: Optional[int]
    anchor_comment_id: Optional[int]




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Sequence
if TYPE_CHECKING:
    from .more_comments_SYNC import MoreComments

from ..util.tree_node import GeneralTreeNode
from .comment_SYNC import Comment
from .submission_SYNC import Submission

T = TypeVar('T')

class ICommentSubtreeTreeNode:
    pass

class CommentSubtreeTreeNode(ICommentSubtreeTreeNode, GeneralTreeNode[T, 'CommentTreeNode']):
    def __init__(self,
        value: T,
        children: Sequence[CommentTreeNode],
        more: Optional[MoreComments],
    ):
        super().__init__(value, children)
        self.more: Optional[MoreComments] = more

class MoreCommentsTreeNode(CommentSubtreeTreeNode[None]):
    pass

class CommentTreeNode(CommentSubtreeTreeNode[Comment]):
    pass

class SubmissionTreeNode(CommentSubtreeTreeNode[Submission]):
    pass




from __future__ import annotations
from typing import Mapping

from http import HTTPStatus

from ..exceptions import ArgExcMixin

class ArgExc(ArgExcMixin):
    pass


class TransportError(ArgExc):
    pass

class TimeoutException(ArgExc):
    pass

class StatusCodeException(ArgExc):
    STATUS_CODE: int = 0

    def __init__(self, arg: object = None, *, status_code: int) -> None:
        super().__init__(arg)
        self.status_code: int = status_code

    def get_default_message(self) -> str:
        sts = self.status_code
        try:
            return f"{sts} {HTTPStatus(sts).phrase}"
        except ValueError:
            return f"{sts}"

class StatusCodeExceptionTypes:
    class InformationalResponseException(StatusCodeException):
        STATUS_CODE: int = -100
    class SuccessfulResponseException(StatusCodeException):
        STATUS_CODE: int = -200
    class RedirectionResponseException(StatusCodeException):
        STATUS_CODE: int = -300
    class ClientErrorResponseException(StatusCodeException):
        STATUS_CODE: int = -400
    class ServerErrorResponseException(StatusCodeException):
        STATUS_CODE: int = -500

    class BadRequest(ClientErrorResponseException):
        STATUS_CODE: int = 400
    class Unauthorized(ClientErrorResponseException):
        STATUS_CODE: int = 401
    class Forbidden(ClientErrorResponseException):
        STATUS_CODE: int = 403
    class NotFound(ClientErrorResponseException):
        STATUS_CODE: int = 404
    class Conflict(ClientErrorResponseException):
        STATUS_CODE: int = 409
    class PayloadTooLarge(ClientErrorResponseException):
        STATUS_CODE: int = 413
    class URITooLong(ClientErrorResponseException):
        STATUS_CODE: int = 414
    class TooManyRequests(ClientErrorResponseException):
        STATUS_CODE: int = 429

    class InternalServerError(ServerErrorResponseException):
        STATUS_CODE: int = 500
    class BadGateway(ServerErrorResponseException):
        STATUS_CODE: int = 502
    class ServiceUnavailable(ServerErrorResponseException):
        STATUS_CODE: int = 503
    class GatewayTimeout(ServerErrorResponseException):
        STATUS_CODE: int = 504

status_code_exception_class_by_status_code: Mapping[int, type[StatusCodeException]] = {
    cls.STATUS_CODE: cls
    for cls in [
        # Client errors
        StatusCodeExceptionTypes.BadRequest,
        StatusCodeExceptionTypes.Unauthorized,
        StatusCodeExceptionTypes.Forbidden,
        StatusCodeExceptionTypes.NotFound,
        StatusCodeExceptionTypes.Conflict,
        StatusCodeExceptionTypes.PayloadTooLarge,
        StatusCodeExceptionTypes.URITooLong,
        StatusCodeExceptionTypes.TooManyRequests,

        # Server errors
        StatusCodeExceptionTypes.InternalServerError,
        StatusCodeExceptionTypes.BadGateway,
        StatusCodeExceptionTypes.ServiceUnavailable,
        StatusCodeExceptionTypes.GatewayTimeout,
    ]
}

def get_status_code_exception_class_by_status_code(n: int) -> type[StatusCodeException]:
    klass = status_code_exception_class_by_status_code.get(n)
    if klass is None:
        klass = StatusCodeException
        if 100 <= n <= 199:
            klass = StatusCodeExceptionTypes.InformationalResponseException
        elif 200 <= n <= 299:
            klass = StatusCodeExceptionTypes.SuccessfulResponseException
        elif 300 <= n <= 399:
            klass = StatusCodeExceptionTypes.RedirectionResponseException
        elif 400 <= n <= 499:
            klass = StatusCodeExceptionTypes.ClientErrorResponseException
        elif 500 <= n <= 599:
            klass = StatusCodeExceptionTypes.ServerErrorResponseException
    return klass

def status_successful(n: int) -> bool:
    return 200 <= n <= 299

def raise_now(n: int) -> None:
    raise get_status_code_exception_class_by_status_code(n)(status_code=n)

def raise_for_status(n: int) -> None:
    if not status_successful(n):
        raise_now(n)




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar
if TYPE_CHECKING:
    from typing import Any, Optional, Mapping, MutableMapping, Union
    from types import TracebackType
    from .session_base_SYNC import SessionBase
    from .requestor_SYNC import Requestor
    from .request import Request
    from .response import Response
    from .payload import RequestFiles

from urllib.parse import urljoin

from .request import make_request
from .util.case_insensitive_dict import CaseInsensitiveDict

T = TypeVar('T')

class HTTPClient:
    @staticmethod
    def make_request(
        verb: str,
        uri: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: Any = None,
        files: Optional[RequestFiles] = None,
    ) -> Request:
        return make_request(
            verb,
            uri,
            params=params,
            headers=headers,
            data=data,
            json=json,
            files=files,
        )

    def __init__(self,
        session: SessionBase,
        requestor: Optional[Requestor] = None,
    ) -> None:
        self.session: SessionBase = session
        self.requestor: Requestor = session if requestor is None else requestor

    def __enter__(self: T) -> T:
        return self

    def __exit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.close()
        return None

    def close(self) -> None:
        self.session.close()

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        return self.requestor.send(request, timeout=timeout)

    def request(self,
        verb: str,
        uri: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: Any = None,
        files: Optional[RequestFiles] = None,
        timeout: float = -2,
    ) -> Response:
        r = self.make_request(verb, uri, params=params, headers=headers,
                data=data, json=json, files=files)
        return self.send(r, timeout=timeout)


class BasicRequestDefaultsHTTPClient(HTTPClient):
    def __init__(self,
        session: SessionBase,
        requestor: Optional[Requestor] = None,
        *,
        params: Optional[MutableMapping[str, str]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
    ) -> None:
        super().__init__(session, requestor)
        self.base_url: str = ''
        self.params: MutableMapping[str, str] = CaseInsensitiveDict() if params is None else params
        self.headers: MutableMapping[str, str] = CaseInsensitiveDict() if headers is None else headers

    def _prepare_request(self, request: Request) -> None:
        r = request
        r.uri = urljoin(self.base_url, r.uri)
        (_d0 := r.params).update({**self.params, **_d0})
        (_d1 := r.headers).update({**self.headers, **_d1})

    def _do_send(self, request: Request, *, timeout: float = -2) -> Response:
        return super().send(request, timeout=timeout)

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        self._prepare_request(request)
        return self._do_send(request, timeout=timeout)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .submission.SYNC import SubmissionProcedures
from .comment.SYNC import CommentProcedures
from .subreddit.SYNC import SubredditProcedures
from .comment_tree.SYNC import CommentTreeProcedures
from .front.SYNC import FrontProcedures
from .user.SYNC import UserProcedures
from .account.SYNC import AccountProcedures
from .collection.SYNC import CollectionProcedures
from .flair.SYNC import FlairProcedures
from .flair_emoji.SYNC import FlairEmojiProcedures
from .custom_feed.SYNC import CustomFeedProcedures
from .live_thread.SYNC import LiveThreadProcedures
from .message.SYNC import MessageProcedures
from .moderation.SYNC import ModerationProcedures
from .draft.SYNC import DraftProcedures
from .misc.SYNC import MiscProcedures
from .modmail.SYNC import ModmailProcedures
from .widget.SYNC import WidgetProcedures
from .wiki.SYNC import WikiProcedures
from .legacy_reddit_subreddit_style.SYNC import LegacyRedditSubredditStyleProcedures
from .redesign_reddit_subreddit_style.SYNC import RedesignRedditSubredditStyleProcedures

class SiteProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.submission: SubmissionProcedures = SubmissionProcedures(client)
        self.comment: CommentProcedures = CommentProcedures(client)
        self.subreddit: SubredditProcedures = SubredditProcedures(client)
        self.comment_tree: CommentTreeProcedures = CommentTreeProcedures(client)
        self.front: FrontProcedures = FrontProcedures(client)
        self.user: UserProcedures = UserProcedures(client)
        self.account: AccountProcedures = AccountProcedures(client)
        self.collection: CollectionProcedures = CollectionProcedures(client)
        self.flair: FlairProcedures = FlairProcedures(client)
        self.flair_emoji: FlairEmojiProcedures = FlairEmojiProcedures(client)
        self.custom_feed: CustomFeedProcedures = CustomFeedProcedures(client)
        self.live_thread: LiveThreadProcedures = LiveThreadProcedures(client)
        self.message: MessageProcedures = MessageProcedures(client)
        self.moderation: ModerationProcedures = ModerationProcedures(client)
        self.draft: DraftProcedures = DraftProcedures(client)
        self.misc: MiscProcedures = MiscProcedures(client)
        self.modmail: ModmailProcedures = ModmailProcedures(client)
        self.widget: WidgetProcedures = WidgetProcedures(client)
        self.wiki: WikiProcedures = WikiProcedures(client)
        self.legacy_reddit_subreddit_style: LegacyRedditSubredditStyleProcedures = LegacyRedditSubredditStyleProcedures(client)
        self.redesign_reddit_subreddit_style: RedesignRedditSubredditStyleProcedures = RedesignRedditSubredditStyleProcedures(client)




from __future__ import annotations

from ..exceptions import ArgExcMixin

class ArgExc(ArgExcMixin):
    pass


class TransportError(ArgExc):
    pass

class TimeoutException(ArgExc):
    pass


class MessageTypeMismatchException(ArgExc):
    pass

class ProtocolViolationException(ArgExc):
    pass

class InvalidStateException(ArgExc):
    pass

class ConnectionClosedException(ArgExc):
    pass




from typing import Callable

import time

class TokenBucket:
    def __init__(self, capacity: float, rate: float, *,
            time_func: Callable[[], float] = time.monotonic) -> None:
        self.capacity: float = capacity
        self.rate: float = rate
        self.time_func: Callable[[], float] = time_func
        self._value = capacity
        self._timestamp = time_func()

    def _checkpoint(self) -> float:
        now = self.time_func()
        delta = now - self._timestamp
        self._timestamp = now
        return delta

    def _replenish(self) -> None:
        if self._value < self.capacity:
            accession = self.rate * self._checkpoint()
            self._value = min(self._value + accession, self.capacity)

    def get_value(self) -> float:
        """Return the number of tokens in the bucket."""
        self._replenish()
        return self._value

    def can_consume(self, n: float) -> bool:
        """Check if `n` tokens are available."""
        return n <= self.get_value()

    def try_consume(self, n: float) -> bool:
        """Consume `n` tokens if `n` tokens are available."""
        if self.can_consume(n):
            self._value -= n
            return True
        return False

    def consume(self, n: float) -> None:
        """Consume `n` tokens. If `n` tokens aren't available, raise an exception."""
        if self.try_consume(n):
            return
        raise RuntimeError(f'attempted to consume {n} tokens when {self._value} were available')

    def hard_consume(self, n: float) -> bool:
        """Consume up to `n` tokens."""
        u = self.get_value()
        self._value = max(u - n, 0)
        return n <= u

    def consume_all(self) -> None:
        """Empty the token bucket completely."""
        self._checkpoint()
        self._value = 0

    def get_cooldown(self, n: float) -> float:
        """Return the duration the client should wait before the consume
        methods return `True` again.
        """
        return max((n - self.get_value()) / self.rate, 0)




from ..carriers.main_sync.makers.subreddit import (  # noqa: F401
    make_submission_stream as make_submission_stream,
    make_comment_stream as make_comment_stream,
    make_subreddit_stream as make_subreddit_stream,
)




from __future__ import annotations
from typing import TypeVar, AsyncIterator, Callable, Optional, Awaitable, Iterable

T = TypeVar('T')

class StubbornCallerAsyncIterator(AsyncIterator[T]):
    def __init__(self, iterable: Iterable[Callable[[], Awaitable[T]]]) -> None:
        self._itr = iter(iterable)
        self.current: Optional[Callable[[], Awaitable[T]]] = None

    def __aiter__(self) -> AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        if self.current is None:
            try:
                self.current = next(self._itr)
            except StopIteration:
                raise StopAsyncIteration
        result = await self.current()
        self.current = None
        return result




from typing import Type, TypeVar
from dataclasses import dataclass

@dataclass
class Frame:
    opcode: int
    fin: bool
    data: bytes

    T = TypeVar('T', bound='Frame')

    @classmethod
    def make(cls: Type[T], opcode: int, data: bytes, fin: bool = True) -> T:
        return cls(opcode=opcode, data=data, fin=fin)


@dataclass
class Message:
    pass

@dataclass
class TextMessage(Message):
    data: str

@dataclass
class BytesMessage(Message):
    data: bytes


@dataclass
class Signal:
    pass

@dataclass
class ConnectionClosed(Signal):
    pass




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from .listing.listing_async_paginator import ListingAsyncPaginator
from ...models.wiki_ASYNC import WikiPageRevision
from ...model_loaders.wiki_ASYNC import load_wiki_page_revision

class WikiPageRevisionsAsyncPaginator(ListingAsyncPaginator[WikiPageRevision]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
    ):
        cursor_extractor = lambda x: ('WikiRevision_' + x['id'])
        super().__init__(client, uri, limit=limit, cursor_extractor=cursor_extractor)

    async def fetch(self) -> Sequence[WikiPageRevision]:
        data = await self._fetch_data()
        return [load_wiki_page_revision(d, self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.message_ASYNC import ComposedMessage

from functools import cached_property

from ...util.base_conversion import to_base36
from ...model_loaders.message_ASYNC import load_composed_message, load_composed_message_thread
from .pull_ASYNC import Pull

class MessageProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.pull: Pull = Pull(client)

    async def get_message_thread(self, idn: int) -> Sequence[ComposedMessage]:
        id36 = to_base36(idn)
        root = await self._client.request('GET', f'/message/messages/{id36}')
        obj_data = root['data']['children'][0]['data']
        return load_composed_message_thread(obj_data, self._client)

    async def send(self, to: str, subject: str, body: str) -> None:
        req_data = {
            'to': to,
            'subject': subject,
            'text': body,
        }
        await self._client.request('POST', '/api/compose', data=req_data)

    async def send_from_sr(self, sr: str, to: str, subject: str, body: str) -> None:
        req_data = {
            'from_sr': sr,
            'to': to,
            'subject': subject,
            'text': body,
        }
        await self._client.request('POST', '/api/compose', data=req_data)

    async def reply(self, idn: int, body: str) -> ComposedMessage:
        data = {
            'thing_id': 't4_' + to_base36(idn),
            'text': body,
            'return_rtjson': '1',
        }
        result = await self._client.request('POST', '/api/comment', data=data)
        root = result['json']['data']['things'][0]['data']
        return load_composed_message(root, self._client)

    async def delete(self, idn: int) -> None:
        await self._client.request('POST', '/api/del_msg', data={'id': 't4_' + to_base36(idn)})

    async def mark_read(self, idn: int) -> None:
        await self._client.request('POST', '/api/read_message', data={'id': 't4_' + to_base36(idn)})

    async def mark_unread(self, idn: int) -> None:
        await self._client.request('POST', '/api/unread_message', data={'id': 't4_' + to_base36(idn)})

    async def mark_all_read(self) -> None:
        await self._client.request('POST', '/api/read_all_messages')

    async def mark_comment_read(self, idn: int) -> None:
        await self._client.request('POST', '/api/read_message', data={'id': 't1_' + to_base36(idn)})

    async def mark_comment_unread(self, idn: int) -> None:
        await self._client.request('POST', '/api/read_message', data={'id': 't1_' + to_base36(idn)})

    async def collapse(self, idn: int) -> None:
        await self._client.request('POST', '/api/collapse_message', data={'id': 't4_' + to_base36(idn)})

    async def uncollapse(self, idn: int) -> None:
        await self._client.request('POST', '/api/uncollapse_message', data={'id': 't4_' + to_base36(idn)})

    class _block_author:
        def __init__(self, outer: MessageProcedures) -> None:
            self._client = outer._client

        async def __call__(self, idn: int) -> None:
            await self.of_message(idn)

        async def of_message(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't4_' + to_base36(idn)})

        async def of_comment(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't1_' + to_base36(idn)})

        async def of_submission(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't3_' + to_base36(idn)})

    block_author: cached_property[_block_author] = cached_property(_block_author)




from __future__ import annotations
from typing import Any, Mapping, Optional

from ..models.flair import (
    FlairTemplate,
    FlairChoices,
    CurrentFlairChoice,
    FlairTemplateChoice,
    UserFlairAssociation,
)

def load_variant1_flair_template(d: Mapping[str, Any]) -> FlairTemplate:
    return FlairTemplate(
        d=d,
        uuid=d['id'],
        type=d['type'],
        mod_only=d['mod_only'],
        text=d['text'],
        bg_color=d['background_color'],
        fg_light_or_dark=d['text_color'],
        allowable_content=d['allowable_content'],
        css_class=d['css_class'],
        max_emojis=d['max_emojis'],
        text_editable=d['text_editable'],
    )

def load_variant2_flair_template(d: Mapping[str, Any]) -> FlairTemplate:
    return FlairTemplate(
        d=d,
        uuid=d['id'],
        type=d['type'],
        mod_only=d['modOnly'],
        text=d['text'],
        bg_color=d['backgroundColor'],
        fg_light_or_dark=d['textColor'],
        allowable_content=d['allowableContent'],
        css_class=d['cssClass'],
        max_emojis=d['maxEmojis'],
        text_editable=d['textEditable'],
    )

def load_flair_template_choice(d: Mapping[str, Any]) -> FlairTemplateChoice:
    return FlairTemplateChoice(
        uuid=d['flair_template_id'],
        text=d['flair_text'],
        css_class=d['flair_css_class'],
        position=d['flair_position'],
        text_editable=d['flair_text_editable'],
    )

def load_current_flair_choice(d: Mapping[str, Any]) -> CurrentFlairChoice:
    return CurrentFlairChoice(
        uuid=d['flair_template_id'],
        text=d['flair_text'],
        css_class=d['flair_css_class'],
        position=d['flair_position'],
    )

def load_flair_choices(d: Mapping[str, Any]) -> FlairChoices:
    current_data = d['current']
    current = None
    if 'flair_template_id' in current_data or 'flair_text' in current_data:
        current = load_current_flair_choice(current_data)

    choices_data = d['choices']
    choices = [load_flair_template_choice(m) for m in choices_data]

    return FlairChoices(current, choices)

def load_user_flair_association(d: Mapping[str, Any]) -> UserFlairAssociation:
    flair_css_class_temp: Optional[str] = d['flair_css_class']
    return UserFlairAssociation(
        username=d['user'],
        text=d['flair_text'],
        has_had_css_class_when_no_flair_template=flair_css_class_temp is not None,
        css_class=flair_css_class_temp or '',
    )




from __future__ import annotations
from typing import Mapping, Any

from enum import IntEnum
from dataclasses import dataclass, field

from .artifact import IArtifact

class ConversationProgress(IntEnum):
    NEW = 0
    IN_PROGRESS = 1
    ARCHIVED = 2

class ModmailModActionType(IntEnum):
    HIGHLIGHT = 0
    UNHIGHLIGHT = 1
    ARCHIVE = 2
    UNARCHIVE = 3
    MUTE_USER = 5
    UNMUTE_USER = 6
    BAN_USER = 7
    UNBAN_USER = 8
    APPROVE_USER = 9
    DISAPPROVE_USER = 10


@dataclass
class ModmailModeratedSubreddit(IArtifact):
    d: Mapping[str, Any] = field(repr=False)
    id: int
    name: str
    subscriber_count: int




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..models.my_account_ASYNC import MyAccount

def load_account(d: Mapping[str, Any], client: Client) -> MyAccount:
    return MyAccount(d, client)




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.modmail import ModmailModeratedSubreddit

from ...model_loaders.modmail import load_modmail_moderated_subreddit
from .conversation_ASYNC import ConversationProcedures

class ModmailProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.conversation: ConversationProcedures = ConversationProcedures(client)

    async def get_unread_counts(self) -> Mapping[str, int]:
        return await self._client.request('GET', '/api/mod/conversations/unread/count')

    async def get_moderating_subreddits(self) -> Sequence[ModmailModeratedSubreddit]:
        root = await self._client.request('GET', '/api/mod/conversations/subreddits')
        return [load_modmail_moderated_subreddit(d) for d in root['subreddits'].values()]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from types import ModuleType
    from importlib.machinery import ModuleSpec

import sys
import importlib.util

def load_module_from_spec(spec: ModuleSpec) -> ModuleType:
    if spec.loader is None:
        raise ImportError('spec has no loader')

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


class _LazyImport:
    def __call__(self, name: str, package: Optional[str] = None) -> ModuleType:
        try:
            return sys.modules[name]
        except KeyError:
            spec = importlib.util.find_spec(name, package)
            if spec is None:
                raise ImportError(f'module named {name!r} not found')
            if spec.loader is None:
                raise ImportError('spec has no loader')

            module = importlib.util.module_from_spec(spec)
            loader = importlib.util.LazyLoader(spec.loader)
            loader.exec_module(module)
            sys.modules[spec.name] = module
            return module

    def __mod__(self, other: str) -> None:
        if '.' in other:
            raise ValueError('dotted module name not supported')

        module = self(other)
        caller_frame = sys._getframe(1)  # type: ignore
        caller_frame.f_globals[other] = module

lazy_import: _LazyImport = _LazyImport()




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...http.session_base_SYNC import SessionBase

from ...core.reddit_http_client_SYNC import RedditHTTPClient
from ...core.recorded_SYNC import Recorded
from ...http.transport.SYNC import new_session
from ...http.util.case_insensitive_dict import CaseInsensitiveDict
from ...util.user_agent_SYNC import get_user_agent_from_session
from ...http.http_client_SYNC import HTTPClient
from ..const import PUSHSHIFT_BASE_URL
from ...core.recorded_SYNC import Last
from .rate_limited_SYNC import RateLimited

def build_http_client(
    *,
    session: Optional[SessionBase] = None,
) -> HTTPClient:
    if session is None:
        session = new_session()
    ua = get_user_agent_from_session(session)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    recorder = Recorded(session)
    last = Last(recorder)
    requestor = RateLimited(recorder)
    http = RedditHTTPClient(session, requestor, headers=headers, last=last)
    http.base_url = PUSHSHIFT_BASE_URL
    return http




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Any, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.custom_feed_ASYNC import CustomFeed

import json

from ...model_loaders.custom_feed_ASYNC import load_custom_feed
from ... import exceptions
from ... import http
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_async_iterator import CallChunkCallingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk

class CustomFeedProcedures:
    def __init__(self, client: Client):
        self._client = client
        self._json_encoder = encoder = json.JSONEncoder()
        self._json_encode = encoder.encode

    async def get(self, user: str, feed: str) -> Optional[CustomFeed]:
        try:
            root = await self._client.request('GET', f'/api/multi/user/{user}/m/{feed}')
        except exceptions.RedditError as e:
            if e.codename == 'MULTI_NOT_FOUND':
                return None
            raise
        return load_custom_feed(root['data'], self._client)

    async def list_created(self, user: str = '') -> Sequence[CustomFeed]:
        uri = '/api/multi/mine'
        if user:
            uri = f'/api/multi/user/{user}'

        result = await self._client.request('GET', uri)
        return [load_custom_feed(d['data'], self._client) for d in result]

    async def create(self,
        user: str, feed: str,
        *,
        title: Optional[str] = None, description: Optional[str] = None,
        subreddit_names: Sequence[str] = (), private: bool = False,
    ) -> CustomFeed:
        json_data: dict[str, Any] = {}
        if title is not None:
            json_data['display_name'] = title
        if description is not None:
            json_data['description_md'] = description
        if subreddit_names:
            json_data['subreddits'] = {'name': nm for nm in subreddit_names}
        if not private:
            json_data['visibility'] = 'public'

        json_str = self._json_encode(json_data)
        root = await self._client.request('POST', f'/api/multi/user/{user}/m/{feed}', data={'model': json_str})
        return load_custom_feed(root['data'], self._client)

    async def put(self,
        user: str, feed: str,
        *,
        title: Optional[str] = None, description: Optional[str] = None,
        subreddit_names: Sequence[str] = (), private: bool = False,
    ) -> CustomFeed:
        json_data: dict[str, Any] = {}
        if title is not None:
            json_data['display_name'] = title
        if description is not None:
            json_data['description_md'] = description
        if subreddit_names:
            json_data['subreddits'] = [{'name': nm} for nm in subreddit_names]
        if not private:
            json_data['visibility'] = 'public'

        json_str = self._json_encode(json_data)
        root = await self._client.request('PUT', f'/api/multi/user/{user}/m/{feed}', data={'model': json_str})
        return load_custom_feed(root['data'], self._client)

    async def delete(self, user: str, feed: str) -> None:
        await self._client.request('DELETE', f'/api/multi/user/{user}/m/{feed}')

    async def duplicate(self, from_user: str, from_feed: str, to_user: str, to_feed: str, *,
            title: Optional[str] = None, description: Optional[str] = None) -> CustomFeed:
        data = {
            'from': f'/user/{from_user}/m/{from_feed}',
            'to': f'/user/{to_user}/m/{to_feed}',
        }
        if title is not None:
            data['display_name'] = title
        if description is not None:
            data['description_md'] = description

        root = await self._client.request('POST', '/api/multi/copy', data=data)
        return load_custom_feed(root['data'], self._client)

    async def contains(self, user: str, feed: str, sr_name: str) -> bool:
        try:
            await self._client.request('GET', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 500:
                return False
            raise
        except exceptions.RedditError as e:
            if e.codename == 'SUBREDDIT_NOEXIST':
                return False
            raise
        return True

    async def add_to(self, user: str, feed: str, sr_name: str) -> None:
        json_str = self._json_encode({"name": "abc"})
        await self._client.request('PUT', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}', data={'model': json_str})

    def bulk_add_to(self, user: str, feed: str, sr_names: Iterable[str]) -> CallChunkCallingAsyncIterator[None]:
        async def mass_add_to(sr_names: Sequence[str]) -> None:
            data = {'path': f'/user/{user}/m/{feed}', 'sr_names': (','.join(sr_names))}
            await self._client.request('POST', '/api/multi/add_srs_bulk', data=data)

        return CallChunkCallingAsyncIterator(AsyncCallChunk(mass_add_to, chunk) for chunk in chunked(sr_names, 300))

    async def remove_from(self, user: str, feed: str, sr_name: str) -> None:
        await self._client.request('DELETE', f'/api/multi/user/{user}/m/{feed}/r/{sr_name}')




from __future__ import annotations
from typing import TYPE_CHECKING, Awaitable, AsyncIterator, Optional, Union
if TYPE_CHECKING:
    from ....client_ASYNC import Client

import asyncio
from asyncio.queues import Queue
import inspect


def run(*streams: Union[Awaitable[None], AsyncIterator[float]], client: Optional[Client] = None) -> None:
    awbls = []
    aitrs = []
    for obj in streams:
        if inspect.isawaitable(obj):
            awbls.append(obj)
        else:
            assert isinstance(obj, AsyncIterator)
            aitrs.append(obj)

    async def main() -> None:
        try:
            await asyncio.gather(
                _run_series_async(*aitrs),
                _run_parallel_async(*awbls),
            )
        finally:
            if client is not None:
                await client.close()

    asyncio.run(main())


async def _run_series_async(*streams: AsyncIterator[float], client: Optional[Client] = None) -> None:
    try:
        loop = asyncio.get_running_loop()

        q: Queue[AsyncIterator[float]] = Queue()
        count = 0
        for aitr in streams:
            count += 1
            q.put_nowait(aitr)

        while count > 0:
            aitr = await q.get()

            try:
                t = await aitr.__anext__()
            except StopAsyncIteration:
                count -= 1
            else:
                loop.call_later(t, (lambda: q.put_nowait(aitr)))

            q.task_done()

    finally:
        if client is not None:
            await client.close()

def run_series(*streams: AsyncIterator[float], client: Optional[Client] = None) -> None:
    async def main() -> None:
        await _run_series_async(*streams, client=client)
    asyncio.run(main())


async def _run_parallel_async(*streams: Awaitable[None], client: Optional[Client] = None) -> None:
    try:
        await asyncio.gather(*streams)
    finally:
        if client is not None:
            await client.close()

def run_parallel(*streams: Awaitable[None], client: Optional[Client] = None) -> None:
    async def main() -> None:
        await _run_parallel_async(*streams, client=client)
    asyncio.run(main())




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.user_SYNC import User
    from ...models.moderated_subreddit import ModeratedSubreddit

from .get_user_summary_SYNC import GetUserSummary
from .bulk_fetch_user_summary_SYNC import BulkFetchUserSummary
from .pull_SYNC import Pull
from .pull_user_subreddits_SYNC import PullUserSubreddits
from ...model_loaders.user_SYNC import load_user, load_potentially_suspended_user
from ...model_loaders.moderated_subreddit import load_moderated_subreddit
from ... import http
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.user.sync1 import SearchUsersListingPaginator
from ...exceptions import RejectedResultException

class UserProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.get_user_summary: GetUserSummary = GetUserSummary(client)
        self.bulk_fetch_user_summary: BulkFetchUserSummary = BulkFetchUserSummary(client)
        self.pull: Pull = Pull(client)
        self.pull_user_subreddits: PullUserSubreddits = PullUserSubreddits(client)

    def get_by_name(self, name: str) -> Optional[User]:
        try:
            root = self._client.request('GET', f'/user/{name}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        obj_data = root['data']
        if obj_data.get('is_suspended', False):
            return None
        return load_user(obj_data, self._client)

    def fetch_by_name(self, name: str) -> User:
        root = self._client.request('GET', f'/user/{name}/about')
        obj_data = root['data']
        if obj_data.get('is_suspended', False):
            raise RejectedResultException('user is suspended')
        return load_user(obj_data, self._client)

    def get_potentially_suspended_user_by_name(self, name: str) -> Optional[object]:
        try:
            root = self._client.request('GET', f'/user/{name}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        return load_potentially_suspended_user(root['data'], self._client)

    def fetch_potentially_suspended_user_by_name(self, name: str) -> object:
        root = self._client.request('GET', f'/user/{name}/about')
        return load_potentially_suspended_user(root['data'], self._client)

    def moderated_subreddits(self, user: str) -> Sequence[ModeratedSubreddit]:
        root = self._client.request('GET', f'/user/{user}/moderated_subreddits')
        return [load_moderated_subreddit(d) for d in root['data']]

    def search(self, query: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[SearchUsersListingPaginator, User]:
        p = SearchUsersListingPaginator(self._client, '/users/search', query)
        return ImpartedPaginatorChainingIterator(p, amount)

    def name_exists(self, name: str) -> bool:
        return not self._client.request('GET', '/api/username_available', params={'user': name})




from .call_chunk import CallChunk as CallChunk  # noqa: F401
from .call_chunk_chaining_iterator import CallChunkChainingIterator as CallChunkChainingIterator  # noqa: F401
from .call_chunk_calling_iterator import CallChunkCallingIterator as CallChunkCallingIterator  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.subreddit_SYNC import Subreddit, InaccessibleSubreddit

def load_subreddit(d: Mapping[str, Any], client: Client) -> Subreddit:
    return Subreddit(d, client)

def load_potentially_inaccessible_subreddit(d: Mapping[str, Any], client: Client) -> object:
    if d['public_traffic'] is None:
        return InaccessibleSubreddit(d, client)
    return Subreddit(d, client)




from __future__ import annotations
from typing import Any, Optional, Sequence, Generic, Mapping, TypeVar

from datetime import datetime
from dataclasses import dataclass

from .artifact import IArtifact


class BaseConversation(IArtifact):
    @dataclass(repr=False, eq=False)
    class LegacyMessage:
        id: int
        id36: str
        link: str

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.subject: str = d['subject']
        self.progress: int = d['state']
        self.message_count: int = d['numMessages']
        self.is_auto: bool = d['isAuto']
        self.is_internal: bool = d['isInternal']
        self.is_repliable: bool = d['isRepliable']
        self.is_highlighted: bool = d['isHighlighted']

        self.legacy_message: Optional[BaseConversation.LegacyMessage] = None
        if (x := d['legacyFirstMessageId']) is not None:
            self.legacy_message = self.LegacyMessage(
                id36=x,
                id=int(x, 36),
                link="https://www.reddit.com/message/messages/" + x,
            )

        self.last_user_update: Optional[datetime] = None if (x := d['lastUserUpdate']) is None else datetime.fromisoformat(x)
        self.last_mod_update: Optional[datetime] = None if (x := d['lastModUpdate']) is None else datetime.fromisoformat(x)
        self.last_update: datetime = datetime.fromisoformat(d['lastUpdated'])
        self.last_unread: Optional[datetime] = None if (x := d['lastUnread']) is None else datetime.fromisoformat(x)
        owner = d['owner']
        self.subreddit_name: str = owner['displayName']
        self.subreddit_id: int = int(owner['id'][3:], 36)

class BaseMessage(IArtifact):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        self.id: int = int(d['id'], 36)
        author = d['author']
        self.author_name: str = author['name']
        self.author_id: int = author['id']
        self.body_text: str = d['bodyMarkdown']
        self.body_html: str = d['body']
        self.datetime: datetime = datetime.fromisoformat(d['date'])
        self.is_internal: bool = d['isInternal']

class BaseModmailModAction(IArtifact):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        self.id: int = int(d['id'], 36)
        self.action_type: int = d['actionTypeId']
        self.agent_name: str = d['author']['name']
        self.agent_id: int = d['author']['id']
        self.datetime: datetime = datetime.fromisoformat(d['date'])

class BaseUserDossier(IArtifact):
    class RecentPost:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.permalink: str
            self.title: str
            self.created_at: datetime

    class RecentComment:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.permalink: str
            self.title: str
            self.body: str
            self.created_at: datetime

    class RecentConvo:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.id: int
            self.subject: str
            self.permalink: str

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        self.id: int = int(d['id'][3:], 36)
        self.name: str = d['name']
        self.created_at: datetime = datetime.fromisoformat(d['created'])
        self.is_shadow_banned: bool = d['isShadowBanned']
        self.is_approved: bool = d['approveStatus']['isApproved']
        mute_status = d['muteStatus']
        self.is_muted: bool = mute_status['isMuted']
        self.mute_reason: str = mute_status['reason']
        self.mute_count: int = mute_status['muteCount']
        self.mute_end_datetime: datetime = datetime.min if (x := mute_status['endDate']) is None else datetime.fromisoformat(x)
        ban_status = d['banStatus']
        self.is_banned: bool = ban_status['isBanned']
        self.ban_reason: str = ban_status['reason']
        self.ban_is_permanent: bool = ban_status['isPermanent']
        self.ban_end_datetime: datetime = datetime.min if (x := ban_status['endDate']) is None else datetime.fromisoformat(x)
        self.recent_posts: Sequence[BaseUserDossier.RecentPost] = [self.RecentPost(m) for m in d['recentPosts']]
        self.recent_comments: Sequence[BaseUserDossier.RecentComment] = [self.RecentComment(m) for m in d['recentComments']]
        self.recent_convos: Sequence[BaseUserDossier.RecentConvo] = [self.RecentConvo(m) for m in d['recentConvos']]


TConversation = TypeVar('TConversation', bound=BaseConversation)
TMessage = TypeVar('TMessage', bound=BaseMessage)
TModmailModAction = TypeVar('TModmailModAction', bound=BaseModmailModAction)

class GBaseConversationAggregate(Generic[TConversation, TMessage, TModmailModAction]):
    def __init__(self,
            conversation: TConversation,
            messages: Sequence[TMessage],
            mod_actions: Sequence[TModmailModAction],
            history: Sequence[object]) -> None:
        self.conversation: TConversation = conversation
        self.messages: Sequence[TMessage] = messages
        self.mod_actions: Sequence[TModmailModAction] = mod_actions
        self.history: Sequence[object] = history

TUserDossier = TypeVar('TUserDossier', bound=BaseUserDossier)

class GBaseUserDossierConversationAggregate(
    GBaseConversationAggregate[TConversation, TMessage, TModmailModAction],
    Generic[TConversation, TMessage, TModmailModAction, TUserDossier],
):
    def __init__(self,
            conversation: TConversation,
            messages: Sequence[TMessage],
            mod_actions: Sequence[TModmailModAction],
            history: Sequence[object],
            user_dossier: TUserDossier) -> None:
        super().__init__(conversation, messages, mod_actions, history)
        self.user_dossier: TUserDossier = user_dossier

class GBaseOptionalUserDossierConversationAggregate(
    GBaseConversationAggregate[TConversation, TMessage, TModmailModAction],
    Generic[TConversation, TMessage, TModmailModAction, TUserDossier],
):
    def __init__(self,
            conversation: TConversation,
            messages: Sequence[TMessage],
            mod_actions: Sequence[TModmailModAction],
            history: Sequence[object],
            user_dossier: Optional[TUserDossier]) -> None:
        super().__init__(conversation, messages, mod_actions, history)
        self.user_dossier: Optional[TUserDossier] = user_dossier




from typing import Union, Iterator, Optional, MutableSequence

import time

from .const import Opcode, ConnectionState, Side
from . import exceptions
from . import events
from .events import Frame, Message, BytesMessage, TextMessage
from .utils import parse_close, serialize_close

class WebSocketConnection:
    DEFAULT_WAITTIME: float = 10.
    SIDE: int = Side.NONE

    @property
    def default_timeout(self) -> float:
        return self.default_waittime

    @default_timeout.setter
    def default_timeout(self, value: float) -> None:
        self.default_waittime: float = value

    def __init__(self) -> None:
        self.default_waittime = self.DEFAULT_WAITTIME
        self.state: ConnectionState = ConnectionState.OPEN
        self.close_code: int = -1
        self.close_reason: str = ''
        self.subprotocol: str = ''

    def __iter__(self) -> Iterator[object]:
        yield from self.cycle(-1)

    def _send_frame_impl(self, m: Frame) -> None:
        pass

    def send_frame(self, m: Frame) -> None:
        if self.state == ConnectionState.CLOSED:
            raise exceptions.ConnectionClosedException
        elif self.state == ConnectionState.CLOSING:
            if m.opcode != Opcode.CLOSE:
                raise exceptions.InvalidStateException('cannot send non-close frame in closing state')
        elif self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'cannot send frame in {self.state.name!r} state')

        self._send_frame_impl(m)

    def send_text(self, data: str) -> None:
        self.send_frame(Frame.make(Opcode.TEXT, data.encode()))

    def send_bytes(self, data: bytes) -> None:
        self.send_frame(Frame.make(Opcode.BINARY, data))

    def send(self, data: Union[str, bytes]) -> None:
        if isinstance(data, str):
            self.send_text(data)
        else:
            self.send_bytes(data)

    def _pulse_impl(self, *, timeout: float = -2) -> Iterator[object]:
        return
        yield

    def pulse(self, *, timeout: float = -2) -> Iterator[object]:
        """Process one frame’s worth of incoming data and possibly generate events for it."""
        if self.state == ConnectionState.CLOSED:
            raise exceptions.ConnectionClosedException
        if self.state not in {ConnectionState.OPEN, ConnectionState.CLOSING}:
            raise exceptions.InvalidStateException(f'cannot ingest frame in {self.state.name!r} state')

        yield from self._pulse_impl(timeout=timeout)

    def cycle(self, t: float = -1) -> Iterator[object]:
        """Yield events from `self.pulse()` over `t` seconds."""
        if t == -1:
            while True:
                for event in self.pulse(timeout=-1):
                    yield event
                    if isinstance(event, events.ConnectionClosed):
                        return

        else:
            tv = t
            if tv == -2:
                tv = self.default_waittime
            tn = time.monotonic()
            while tv > 0:
                try:
                    for event in self.pulse(timeout=tv):
                        yield event
                        if isinstance(event, events.ConnectionClosed):
                            return
                except exceptions.TimeoutException:
                    return

                now = time.monotonic()
                tv -= now - tn
                tn = now

    def receive(self, *, timeout: float = -2) -> Message:
        for event in self.cycle(timeout):
            if isinstance(event, Message):
                return event
            elif isinstance(event, events.ConnectionClosed):
                raise exceptions.ConnectionClosedException
        raise exceptions.TimeoutException

    def recv_bytes(self, *, timeout: float = -2) -> bytes:
        """Receive the next message. If it's a BytesMessage then return its payload,
        otherwise raise an MessageTypeMismatchException exception."""
        event = self.receive(timeout=timeout)
        if isinstance(event, BytesMessage):
            return event.data
        raise exceptions.MessageTypeMismatchException('string message received')

    def recv(self, *, timeout: float = -2) -> bytes:
        return self.recv_bytes(timeout=timeout)

    def recv_text(self, *, timeout: float = -2) -> str:
        """Receive the next message. If it's a TextMessage then return its payload,
        otherwise raise an MessageTypeMismatchException exception."""
        event = self.receive(timeout=timeout)
        if isinstance(event, TextMessage):
            return event.data
        raise exceptions.MessageTypeMismatchException('bytes message received')

    def _send_close_frame_impl(self, code: Optional[int] = 1000, reason: str = '') -> None:
        data = b''
        if code is not None:
            data = serialize_close(code, reason)
        close = Frame.make(Opcode.CLOSE, data)
        self.send_frame(close)

    def close(self, code: Optional[int] = 1000, reason: str = '', *, waitfor: float = -2) -> None:
        if code is None and reason:
            raise ValueError('cannot send a reason without a code')

        if self.state == ConnectionState.CLOSED:
            return
        if self.state != ConnectionState.OPEN:
            raise exceptions.InvalidStateException(f'operation not permitted in {self.state.name!r} state')

        self.state = ConnectionState.CLOSING

        self._send_close_frame_impl(code, reason)

        for _ in self.cycle(waitfor):
            pass

        self.shutdown()

    def shutdown(self) -> None:
        if self.close_code < 0:
            self.close_code = 1006
        self.state = ConnectionState.CLOSED



class PulsePartiallyImplementedWebSocketConnection(WebSocketConnection):
    def __init__(self) -> None:
        super().__init__()
        self._accumulator: MutableSequence[Frame] = []

    def _process_ping_frame(self, m: Frame) -> Iterator[object]:
        if self.state not in {ConnectionState.CLOSING, ConnectionState.CLOSED}:
            pong = Frame.make(Opcode.PONG, m.data)
            self.send_frame(pong)
        yield m

    def _process_close_frame(self, m: Frame) -> Iterator[object]:
        self.state: ConnectionState = ConnectionState.CLOSING
        self.close_code: int
        self.close_reason: str
        self.close_code, self.close_reason = parse_close(m.data)

        close = Frame.make(Opcode.CLOSE, m.data)
        try:
            self.send_frame(close)
        except exceptions.ConnectionClosedException:
            pass

        self.shutdown()

        yield m
        yield events.ConnectionClosed()

    def _process_data_frame(self, m: Frame) -> Iterator[object]:
        yield m
        acc = self._accumulator
        acc.append(m)
        if m.fin:
            msg = b''.join(m.data for m in acc)

            if acc[0].opcode & Opcode.TEXT:
                yield TextMessage(msg.decode())
            else:
                yield BytesMessage(msg)

            acc.clear()

    def _process_frame(self, m: Frame) -> Iterator[object]:
        if m.opcode == Opcode.PING:
            yield from self._process_ping_frame(m)
        elif m.opcode == Opcode.CLOSE:
            yield from self._process_close_frame(m)
        elif m.opcode in {Opcode.CONTINUATION, Opcode.TEXT, Opcode.BINARY}:
            yield from self._process_data_frame(m)

    def _load_next_frame(self, *, timeout: float = -2) -> Frame:
        raise NotImplementedError

    def _pulse_impl(self, *, timeout: float = -2) -> Iterator[object]:
        frame = self._load_next_frame(timeout=timeout)
        yield from self._process_frame(frame)




from __future__ import annotations
from typing import Sequence

from .listing_paginator import ListingPaginator
from ....models.comment_SYNC import ExtraSubmissionFieldsComment
from ....model_loaders.comment_SYNC import load_extra_submission_fields_comment

class ExtraSubmissionFieldsCommentListingPaginator(ListingPaginator[ExtraSubmissionFieldsComment]):
    def fetch(self) -> Sequence[ExtraSubmissionFieldsComment]:
        data = self._fetch_data()
        return [load_extra_submission_fields_comment(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional, Sequence, Iterable, Tuple
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ..paginator import CursorPaginator
from ...model_loaders.modmail_SYNC import load_conversation, load_message
from ...models.modmail_SYNC import Conversation, Message

class ModmailConversationsPaginator(CursorPaginator[Tuple[Conversation, Message]]):
    def __init__(self,
        client: Client,
        uri: str,
        mailbox: str,
        subreddit_names: Sequence[str],
        sort: str,
        *,
        limit: Optional[int] = 100,
    ):
        super().__init__()
        self.limit: Optional[int] = limit
        self.client: Client = client
        self.uri: str = uri
        self.mailbox: str = mailbox
        self.subreddit_names: Sequence[str] = subreddit_names
        self.sort: str = sort
        self._after: str = ''
        self._has_after: bool = True

    def get_cursor(self) -> str:
        return self._after

    def set_cursor(self, value: str) -> None:
        self._after = value

    def more_available(self) -> bool:
        return self._has_after

    def set_more_available_flag(self, value: bool) -> None:
        self._has_after = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        if self.limit is not None:
            yield ('limit', str(self.limit))

        if self._after:
            yield ('after', self._after)

        if self.mailbox:
            yield ('state', self.mailbox)
        if self.subreddit_names:
            yield ('entity', ','.join(self.subreddit_names))
        if self.sort:
            yield ('sort', self.sort)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        data = self.client.request('GET', self.uri, params=params)
        entries = data['conversationIds']

        if entries:
            self._after = entries[-1]
        self._has_after = bool(entries)

        return data

    def fetch(self) -> Sequence[Tuple[Conversation, Message]]:
        data = self._fetch_data()
        conversations_mapping = data['conversations']
        messages_mapping = data['messages']
        results = []
        for convo_id36 in data['conversationIds']:
            conversation_data = conversations_mapping[convo_id36]
            message_id36 = conversation_data['objIds'][0]['id']
            message_data = messages_mapping[message_id36]
            results.append(
                (
                    load_conversation(conversation_data, self.client),
                    load_message(message_data, self.client),
                )
            )
        return results




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Iterable
if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

import httpx  # type: ignore[import]

from ..SYNC import register
from ...session_base_SYNC import SessionBase
from ... import exceptions
from ... import payload
from ...response import UResponse

def _generate_request_kwargs(r: Request, etv: float) -> Iterable[tuple[str, Any]]:
    timeout_obj = httpx.Timeout(etv, pool=20)
    if etv == -1:
        timeout_obj = httpx.Timeout(None, pool=20)
    yield ('timeout', timeout_obj)

    yield ('method', r.verb)
    yield ('url', r.uri)
    yield ('params', r.params)

    headers = dict(r.headers)
    pld = r.payload
    if pld is None:
        pass

    elif isinstance(pld, payload.Bytes):
        pld.apply_content_type(headers)
        yield ('data', pld.data)

    elif isinstance(pld, payload.Text):
        pld.apply_content_type(headers)
        yield ('data', pld.text.encode())

    elif isinstance(pld, payload.JSON):
        yield ('json', pld.json)

    elif isinstance(pld, payload.URLEncodedFormData):
        yield ('data', dict(pld.data))

    elif isinstance(pld, payload.MultipartFormData):
        text_plds: list[payload.MultipartTextField] = []
        file_plds: list[payload.MultipartFileField] = []
        for part in pld.parts:
            if isinstance(part, payload.MultipartTextField):
                text_plds.append(part)
            elif isinstance(part, payload.MultipartFileField):
                file_plds.append(part)

        if not file_plds:
            # httpx won't send a multipart if no files
            raise Exception('multipart without file fields not supported')

        yield ('data', {ty.name: ty.value for ty in text_plds})
        yield ('files', {
            fy.name: (fy.filename, fy.file, fy.content_type)
            for fy in file_plds
        })

    else:
        raise Exception('unsupported payload type')

    yield ('headers', headers)


class Session(SessionBase):
    def __init__(self,
        httpx_client: httpx.Client,
    ) -> None:
        super().__init__()
        self.client: httpx.Client = httpx_client

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        etv = self._get_effective_timeout_value(timeout)
        kwargs = dict(_generate_request_kwargs(request, etv))
        try:
            response = self.client.request(**kwargs)
        except httpx.TimeoutException as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        return UResponse(
            status=response.status_code,
            headers=response.headers,
            data=response.content,
            underlying_object=response,
        )

    def close(self) -> None:
        self.client.close()

def new_session() -> Session:
    cl = httpx.Client()
    return Session(cl)

name: str = httpx.__name__
version: str = httpx.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_session=new_session,
)




from __future__ import annotations
from typing import Mapping, Any

from datetime import datetime, timezone

from .artifact import Artifact

class BaseMyAccount(Artifact):
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]):
            self.name: str = d['display_name']
            self.openness: str = d['subreddit_type']
            self.subscriber_count: int = d['subscribers']
            self.title: str = d['title']
            self.public_description: str = d['public_description']
            self.nsfw: bool = d['over_18']

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.name: str = d['name']
        self.has_mail: bool = d['has_mail']
        self.has_mod_mail: bool = d['has_mod_mail']
        self.inbox_count: int = d['inbox_count']
        self.coins: int = d['coins']
        self.friend_count: int = d['num_friends']
        self.is_suspended: bool = d['is_suspended']
        self.nsfw: bool = d['over_18']

        self.subreddit: BaseMyAccount.Subreddit = self.Subreddit(d['subreddit'])




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.submission_draft_SYNC import DraftList
from .submission_draft import load_draft
from .subreddit_SYNC import load_subreddit

def load_draft_list(d: Mapping[str, Any], client: Client) -> DraftList:
    drafts = [load_draft(m) for m in d['drafts']]
    subreddits = [load_subreddit(m, client) for m in d['subreddits']]
    return DraftList(drafts, subreddits)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Optional
if TYPE_CHECKING:
    from ..auth.typedefs import ClientCredentials, AuthorizationGrant
    from ..http.requestor_SYNC import Requestor

from ..http.request import make_request
from ..http.util.json_load import json_loads_response
from ..auth.token_obtainment_client_SYNC import TokenObtainmentClient
from ..auth.utils import apply_basic_auth
from .exceptions import raise_for_reddit_token_server_response_error, raise_for_reddit_auth_response_exception

class RedditTokenObtainmentClient(TokenObtainmentClient):
    def __init__(self, requestor: Requestor, uri: str,
            client_credentials: ClientCredentials,
            grant: AuthorizationGrant,
            headers: Optional[Mapping[str, str]] = None):
        super().__init__(requestor, uri, client_credentials, grant)
        self.headers: Mapping[str, str] = {} if headers is None else headers

    def fetch_data(self) -> Mapping[str, Any]:
        r = make_request('POST', self.uri, data=self.grant)
        apply_basic_auth(r, *self.client_credentials)
        r.headers.update(self.headers)
        resp = self.requestor.send(r)

        try:
            try:
                resp_json = json_loads_response(resp)
            except ValueError as cause:
                try:
                    resp.raise_for_status()
                except Exception as exc:
                    raise exc from cause
                raise

            raise_for_reddit_token_server_response_error(resp_json)
            resp.raise_for_status()

        except Exception as cause:
            try:
                raise_for_reddit_auth_response_exception(cause, r, resp)
            except Exception as exc:
                raise exc from cause
            raise

        return resp_json




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_SYNC import Comment

from ...model_loaders.comment_SYNC import load_comment
from ...util.base_conversion import to_base36
from ...util.extract_id_from_url import extract_comment_id_from_url

class Get:
    def __init__(self, client: Client):
        self._client = client

    def __call__(self, id: int) -> Optional[Comment]:
        id36 = to_base36(id)
        return self.by_id36(id36)

    def by_id36(self, id36: str) -> Optional[Comment]:
        full_id36 = 't1_' + id36
        root = self._client.request('GET', '/api/info', params={'id': full_id36})
        if children := root['data']['children']:
            return load_comment(children[0]['data'], self._client)
        return None

    def by_url(self, url: str) -> Optional[Comment]:
        return self(extract_comment_id_from_url(url))




from __future__ import annotations

import re

RECOMMENDED_BOT_USER_AGENT_PATTERN: str = r'''
^(?:(?P<platform>[\w-]+):)?

(?P<app_id>[\w-]+)

(?:(?(platform)[:/]|/)
(?P<version>[\w\d.-]+))?

(?:\ +\(\ *by\ +(?P<author>\S+).*\))?

(?:\ +.*)?$
'''
RECOMMENDED_BOT_USER_AGENT_REGEX: re.Pattern[str] = re.compile(RECOMMENDED_BOT_USER_AGENT_PATTERN, re.X)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable
if TYPE_CHECKING:
    from ..auth.reddit_internal_api_token_obtainment_client_SYNC import RedditInternalAPITokenObtainmentClient
    from ..auth.token import Token
    from ...http.requestor_SYNC import Requestor
    from ...http.request import Request
    from ...http.response import Response

import time

from ...http.requestor_augmenter_SYNC import RequestorAugmenter

class Authorizer:
    def __init__(self,
        token_client: Optional[RedditInternalAPITokenObtainmentClient] = None,
        token: Optional[Token] = None,
    ):
        self.token_client: Optional[RedditInternalAPITokenObtainmentClient] = token_client
        self.token: Optional[Token] = token
        self.renewal_time: Optional[int] = None
        self.renewal_skew: int = 30
        self.expires_in_fallback: Optional[int] = None
        self.time_func: Callable[[], float] = time.monotonic
        self.authorization_header_name: str = 'Authorization'

    def time(self) -> float:
        return self.time_func()

    def renew_token(self) -> None:
        if self.token_client is None:
            raise RuntimeError('no token client')

        self.token = tk = self.token_client.fetch_token()

        expires_in: Optional[int] = tk.expires_in
        if expires_in is None:
            expires_in = self.expires_in_fallback
        if expires_in is None:
            self.renewal_time = None
        else:
            self.renewal_time = int(self.time()) + expires_in - self.renewal_skew

    def should_renew_token(self) -> bool:
        if self.token is None:
            return True
        if self.renewal_time is None:
            return False
        return self.time() >= self.renewal_time

    def can_renew_token(self) -> bool:
        return self.token_client is not None

    def get_token(self) -> Token:
        if (tk := self.token) is not None:
            return tk
        raise RuntimeError('token not set')

    def set_token(self, token: Token) -> None:
        self.token = token

    def set_access_token(self, access_token: str) -> None:
        self.set_token(Token(access_token))

    def prepare_request(self, request: Request, token: Optional[Token] = None) -> None:
        tk = self.get_token() if token is None else token
        request.headers[self.authorization_header_name] = f'bearer {tk.access_token}'


class Authorized(RequestorAugmenter):
    """Used to perform requests to endpoints that require authorization."""

    def __init__(self, requestor: Requestor, authorizer: Authorizer) -> None:
        super().__init__(requestor)
        self.authorizer: Authorizer = authorizer

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        authorizer = self.authorizer
        if authorizer.should_renew_token():
            authorizer.renew_token()

        authorizer.prepare_request(request)

        resp = self.requestor.send(request, timeout=timeout)

        maybe_invalid_token = resp.status == 401
        if maybe_invalid_token and authorizer.can_renew_token():
            authorizer.renew_token()

            authorizer.prepare_request(request)

            resp = self.requestor.send(request, timeout=timeout)

        return resp




from __future__ import annotations
from typing import Any, Mapping, Collection

from functools import cached_property
from datetime import datetime, timezone

class ModeratorListItem:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        self.id: int = int(id36, 36)
        self.name: str = d['name']
        #self.rel_id: str = d['rel_id']
        self.added_ut: int = int(d['date'])
        self.flair_text: str = d['author_flair_text']
        self.flair_css_class: str = d['author_flair_css_class']
        self.permissions: Collection[str] = d['mod_permissions']

    @cached_property
    def added_at(self) -> datetime:
        return datetime.fromtimestamp(self.added_ut, timezone.utc)




from __future__ import annotations
from typing import Mapping, Any, Optional, Sequence, Generic, TypeVar

from datetime import datetime, timezone

from ..auth.const import AUTHORIZATION_BASE_URL
from .artifact import Artifact
from .report import ModReport, UserReport
from ..model_loaders.report import load_mod_report, load_user_report

from dataclasses import dataclass

class BaseSubmission(Artifact):
    class Me:
        def __init__(self, d: Mapping[str, Any]):
            self.saved: bool = d['saved']
            self.hidden: bool = d['hidden']
            self.reply_notifications: bool = d['send_replies']
            self.voted: int = {False: -1, None: 0, True: 1}[d['likes']]
            self.is_following_event: bool = d.get('is_followed', False)

    class Author:
        class AuthorFlair:
            def __init__(self, d: Mapping[str, Any]):
                self.template_uuid: Optional[str] = d['author_flair_template_id']
                author_flair_text: Optional[str] = d['author_flair_text']
                self.text: str = author_flair_text or ''
                self.has_had_flair: bool = author_flair_text is not None
                self.fg_light_or_dark: str = d['author_flair_text_color'] or ''
                self.bg_color: str = d['author_flair_background_color'] or ''
                self.uses_richtext: bool = d['author_flair_type'] == 'richtext'
                author_flair_css_class: Optional[str] = d['author_flair_css_class']
                self.has_had_css_class_when_no_flair_template: bool = author_flair_css_class is not None
                self.css_class: str = author_flair_css_class or ''

        def __init__(self, d: Mapping[str, Any]):
            self.name: str = d['author']
            self.id36: str = d['author_fullname'].split('_', 1)[-1]
            self.id: int = int(self.id36, 36)
            self.has_premium: bool = d['author_premium']
            self.flair: BaseSubmission.Author.AuthorFlair = self.AuthorFlair(d)

    class Subreddit:
        def __init__(self, d: Mapping[str, Any]):
            self.id36: str = d['subreddit_id'].split('_', 1)[-1]
            self.id: int = int(self.id36, 36)
            self.name: str = d['subreddit']
            self.openness: str = d['subreddit_type']
            self.quarantined: bool = d['quarantine']
            self.subscriber_count: int = d['subreddit_subscribers']

    class Moderator:
        class Approved:
            def __init__(self, d: Mapping[str, Any]):
                self.by: str = d['approved_by']
                self.ut: int = d['approved_at_utc']
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)

        class Removed:
            def __init__(self, d: Mapping[str, Any]):
                self.by: str = d['banned_by']
                self.ut: int = d['banned_at_utc']
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)

        class Reports:
            def __init__(self, d: Mapping[str, Any]):
                self.ignoring: bool = d['ignore_reports']
                self.num_reports: int = d['num_reports']
                self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
                self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]

        def __init__(self, d: Mapping[str, Any]):
            self.spam: bool = d['spam']

            self.approved: Optional[BaseSubmission.Moderator.Approved] = None
            if d['approved_by']:
                self.approved = self.Approved(d)

            self.removed: Optional[BaseSubmission.Moderator.Removed] = None
            if d['banned_by']:
                self.removed = self.Removed(d)

            self.reports: BaseSubmission.Moderator.Reports = self.Reports(d)

            self.removal_reason_by: Optional[str] = d['mod_reason_by']
            self.removal_reason_title: Optional[str] = d['mod_reason_title']
            self.removal_note: Optional[str] = d['mod_note']

    class Event:
        def __init__(self, d: Mapping[str, Any]):
            self.start_ut: int = int(d['event_start'])
            self.start_at: datetime = datetime.fromtimestamp(self.start_ut, timezone.utc)
            self.end_ut: int = int(d['event_end'])
            self.end_at: datetime = datetime.fromtimestamp(self.end_ut, timezone.utc)
            self.is_live: bool = d['event_is_live']

    class Flair:
        def __init__(self, d: Mapping[str, Any]):
            self.has_flair: bool = d['link_flair_text'] is not None
            self.bg_color: str = d['link_flair_background_color']
            link_flair_css_class_temp: Optional[str] = d['link_flair_css_class']
            self.css_class: str = link_flair_css_class_temp or ''
            self.template_uuid: Optional[str] = d.get('link_flair_template_id', None)
            self.text: str = d['link_flair_text'] or ''
            self.fg_light_or_dark: str = d['link_flair_text_color']
            self.type: str = d['link_flair_type']

    class Reports:
        def __init__(self, d: Mapping[str, Any]):
            self.ignoring: bool = d['ignore_reports']
            self.num_reports: int = d['num_reports']
            self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
            self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)

        self.title: str = d['title']
        self.score: int = d['score']
        self.score_hidden: bool = d['hide_score']
        self.comment_count: int = d['num_comments']

        self.rel_permalink: str = d['permalink']
        self.permalink: str = AUTHORIZATION_BASE_URL + d['permalink']

        a: Any = d['edited']
        self.edited: bool = bool(a)
        self.edited_ut: Optional[int] = int(a) if self.edited else None
        self.edited_at: Optional[datetime] = None
        if self.edited_ut is not None:
            self.edited_at = datetime.fromtimestamp(self.edited_ut, timezone.utc)

        self.upvote_ratio: float = d['upvote_ratio']
        self.removal_category: Optional[str] = d['removed_by_category']
        self.suggested_sort: Optional[str] = d['suggested_sort']
        self.stickied: bool = d['stickied']
        self.archived: bool = d['archived']
        self.locked: bool = d['locked']
        self.in_contest_mode: bool = d['contest_mode']
        self.nsfw: bool = d['over_18']
        self.spoiler: bool = d['spoiler']
        self.num_crossposts: int = d['num_crossposts']
        self.is_crosspostable: bool = d['is_crosspostable']
        self.is_original_content: bool = d['is_original_content']
        self.is_robot_indexable: bool = d['is_robot_indexable']
        self.pinned: bool = d['pinned']
        self.distinguished: str = d['distinguished'] or ''

        self.event: Optional[BaseSubmission.Event] = None
        if 'event_start' in d:
            self.event = self.Event(d)

        self.me: BaseSubmission.Me = self.Me(d)

        self.subreddit: BaseSubmission.Subreddit = self.Subreddit(d)

        author: str = d['author']
        self.author_name: str = author
        self.author: Optional[BaseSubmission.Author] = None
        if not author.startswith('['):
            self.author = self.Author(d)

        self.mod: Optional[BaseSubmission.Moderator] = None
        if 'spam' in d:
            self.mod = self.Moderator(d)

        self.flair: BaseSubmission.Flair = self.Flair(d)

        self.reports: Optional[BaseSubmission.Reports] = None
        if d['num_reports'] is not None:
            self.reports = self.Reports(d)


class BaseLinkPost(BaseSubmission):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.link: str = d['url_overridden_by_dest']

class BaseTextPost(BaseSubmission):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.body: str = d['selftext']
        self.body_html: str = d['selftext_html']

class BaseGalleryPost(BaseSubmission):
    @dataclass(repr=False, eq=False)
    class GalleryItem:
        id: int
        media_id: str
        caption: str
        outbound_link: str

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.gallery_link: str = d['url_overridden_by_dest']

        gallery_data_items: Sequence[Any] = ()
        if gallery_data := d.get('gallery_data'):
            gallery_data_items = gallery_data['items']
        self.gallery: Sequence[BaseGalleryPost.GalleryItem] = [
            self.GalleryItem(
                id=m['id'],
                media_id=m['media_id'],
                caption=m.get('caption', ''),
                outbound_link=m.get('outbound_url', ''),
            )
            for m in gallery_data_items
        ]

class BasePollPost(BaseSubmission):
    pass


TOriginalSubmission = TypeVar('TOriginalSubmission', bound=BaseSubmission)

class GBaseCrosspostSubmission(BaseSubmission, Generic[TOriginalSubmission]):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.original_id36: str = d['crosspost_parent'][3:]
        self.original_id: int = int(self.original_id36, 36)
        self.original: TOriginalSubmission = self._load_submission(d['crosspost_parent_list'][0])

    def _load_submission(self, d: Mapping[str, Any]) -> TOriginalSubmission:
        raise NotImplementedError




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableMapping
if TYPE_CHECKING:
    from ...http.session_base_SYNC import SessionBase
    from .authorizer_SYNC import Authorizer
    from ...core.recorded_SYNC import Last
    from ...http.requestor_SYNC import Requestor

from ...core.reddit_http_client_SYNC import RedditHTTPClient
from ...core.rate_limited_SYNC import RateLimited
from ...core.recorded_SYNC import Recorded
from ...core.reddit_give_me_json_please_SYNC import RedditGiveMeJSONPlease
from ...http.misc.apply_params_and_headers_SYNC import ApplyDefaultHeaders
from ...http.transport.SYNC import new_session
from ...http.util.case_insensitive_dict import CaseInsensitiveDict
from ..auth.reddit_internal_api_token_obtainment_client_SYNC import new_reddit_internal_api_token_obtainment_client
from .authorizer_SYNC import Authorized
from ...util.user_agent_SYNC import get_user_agent_from_session


class InternalAPIRedditHTTPClient(RedditHTTPClient):
    @property
    def authorizer(self) -> Authorizer:
        return self.get_authorizer()

    def __init__(self,
        session: SessionBase,
        requestor: Optional[Requestor] = None,
        *,
        params: Optional[MutableMapping[str, str]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
        last: Optional[Last] = None,
        authorizer: Optional[Authorizer] = None,
    ) -> None:
        super().__init__(session, requestor, params=params, headers=headers, last=last)
        self._authorizer: Optional[Authorizer] = authorizer

    def get_authorizer(self) -> Authorizer:
        if self._authorizer is None:
            raise RuntimeError('value not set')
        return self._authorizer

    def set_authorizer(self, value: Authorizer) -> None:
        self._authorizer = value


def build_internal_api_reddit_http_client(
    *,
    session: Optional[SessionBase] = None,
) -> InternalAPIRedditHTTPClient:
    if session is None:
        session = new_session()
    ua = get_user_agent_from_session(session)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    recorder = Recorded(session)
    last = Last(recorder)
    token_client = new_reddit_internal_api_token_obtainment_client(ApplyDefaultHeaders(session, headers))
    authorizer = Authorizer(token_client)
    requestor = RedditGiveMeJSONPlease(RateLimited(Authorized(recorder, authorizer)))
    http = InternalAPIRedditHTTPClient(session, requestor, headers=headers, authorizer=authorizer, last=last)
    return http




from __future__ import annotations

import struct

from . import const
from . import exceptions

def parse_close(data: bytes) -> tuple[int, str]:
    length = len(data)
    if length > 1:
        code, = struct.unpack("!H", data[:2])
        reason = data[2:].decode()
        return (code, reason)
    elif length == 0:
        return (1005, '')
    raise ValueError('close frame payload is too short')

def serialize_close(code: int, reason: str) -> bytes:
    return struct.pack("!H", code) + reason.encode()

def check_close_frame_close_code(code: int) -> None:
    codes = const.AUTHORITATIVE_CLOSE_FRAME_CLOSE_CODES
    if not (code in codes or 3000 <= code < 5000):
        raise exceptions.ProtocolViolationException('invalid close frame close code')




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Iterable, Callable, Any
if TYPE_CHECKING:
    from ....client_ASYNC import Client

from ..listing.listing_async_paginator import ListingAsyncPaginator
from ..listing.mixins.subreddit_detail_ASYNC import SubredditDetail
from ..listing.submission_and_comment_listing_async_paginator import SubmissionAndExtraSubmissionFieldsCommentListingAsyncPaginator
from ..listing.submission_listing_async_paginator import SubmissionListingAsyncPaginator
from ..listing.comment_listing_async_paginator import ExtraSubmissionFieldsCommentListingAsyncPaginator
from ....models.submission_ASYNC import Submission
from ....models.comment_ASYNC import Comment

T = TypeVar('T')

class OnlyFilter(ListingAsyncPaginator[T]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ):
        super().__init__(client, uri, limit=limit, cursor_extractor=cursor_extractor)
        self._only = ''

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        if self._only:
            yield ('only', self._only)

class SubmissionOnlyFilter(OnlyFilter[Submission]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ):
        super().__init__(client, uri, limit=limit, cursor_extractor=cursor_extractor)
        self._only = 'links'

class CommentOnlyFilter(OnlyFilter[Comment]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ):
        super().__init__(client, uri, limit=limit, cursor_extractor=cursor_extractor)
        self._only = 'comments'


class ModQueueListingAsyncPaginator(
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass
class ModQueueSubmissionListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingAsyncPaginator,
): pass
class ModQueueCommentListingAsyncPaginator(
    SubredditDetail[Comment],
    CommentOnlyFilter,
    ExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass

class ReportsListingAsyncPaginator(
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass
class ReportsSubmissionListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingAsyncPaginator,
): pass
class ReportsCommentListingAsyncPaginator(
    SubredditDetail[Comment],
    CommentOnlyFilter,
    ExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass

class SpamListingAsyncPaginator(
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass
class SpamSubmissionListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingAsyncPaginator,
): pass
class SpamCommentListingAsyncPaginator(
    SubredditDetail[Comment],
    CommentOnlyFilter,
    ExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass

class EditedListingAsyncPaginator(
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass
class EditedSubmissionListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingAsyncPaginator,
): pass
class EditedCommentListingAsyncPaginator(
    SubredditDetail[Comment],
    CommentOnlyFilter,
    ExtraSubmissionFieldsCommentListingAsyncPaginator,
): pass

class UnmoderatedSubmissionListingAsyncPaginator(
    SubredditDetail[Submission],
    SubmissionListingAsyncPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.subreddit_SYNC import Subreddit

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.user.sync1 import UserSubredditListingPaginator

class PullUserSubreddits:
    def __init__(self, client: Client):
        self._client = client

    def popular(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingIterator[UserSubredditListingPaginator, Subreddit]:
        p = UserSubredditListingPaginator(self._client, '/users/popular')
        return ImpartedPaginatorChainingIterator(p, amount)

    def new(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingIterator[UserSubredditListingPaginator, Subreddit]:
        p = UserSubredditListingPaginator(self._client, '/users/new')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .message_base import (
    BaseMailboxMessage,
    BaseComposedMessage,
    BaseCommentMessage,
)

class MailboxMessage(BaseMailboxMessage):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

class ComposedMessage(MailboxMessage, BaseComposedMessage):
    def reply(self, body: str) -> ComposedMessage:
        return self.client.p.message.reply(self.id, body)

    def mark_read(self) -> None:
        self.client.p.message.mark_read(self.id)

    def mark_unread(self) -> None:
        self.client.p.message.mark_unread(self.id)

class CommentMessage(MailboxMessage, BaseCommentMessage):
    def mark_read(self) -> None:
        self.client.p.message.mark_comment_read(self.comment.id)

    def mark_unread(self) -> None:
        self.client.p.message.mark_comment_unread(self.comment.id)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union
if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

import json

import urllib3  # type: ignore[import]
import urllib3.exceptions as urllib3_exceptions  # type: ignore[import]

from ..SYNC import register
from ...session_base_SYNC import SessionBase
from ... import exceptions
from ... import payload
from ...response import UResponse
from ...util.merge_query_params import merge_query_params

class Session(SessionBase):
    def __init__(self,
        http: urllib3.poolmanager.PoolManager,
    ) -> None:
        super().__init__()
        self.http: urllib3.poolmanager.PoolManager = http

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        etv = self._get_effective_timeout_value(timeout)
        tmo: Optional[float] = etv
        if tmo == -1:
            tmo = None

        r = request
        url = merge_query_params(r.uri, r.params)
        headers = dict(r.headers)
        pld = r.payload

        response: urllib3.response.HTTPResponse
        try:
            if pld is None:
                response = self.http.urlopen(r.verb, url, headers=headers, timeout=tmo)

            elif isinstance(pld, payload.Bytes):
                pld.apply_content_type(headers)
                response = self.http.urlopen(r.verb, url, headers=headers, body=pld.data, timeout=tmo)

            elif isinstance(pld, payload.Text):
                pld.apply_content_type(headers)
                response = self.http.urlopen(r.verb, url, headers=headers, body=pld.text.encode(), timeout=tmo)

            elif isinstance(pld, payload.JSON):
                pld.apply_content_type(headers)
                response = self.http.urlopen(r.verb, url, headers=headers, body=json.dumps(pld.json).encode(), timeout=tmo)

            elif isinstance(pld, payload.URLEncodedFormData):
                fields0: dict[str, str] = dict(pld.data)
                response = self.http.request_encode_body(r.verb, url, headers=headers, fields=fields0, timeout=tmo,
                        encode_multipart=False)

            elif isinstance(pld, payload.MultipartFormData):
                fields: dict[str, Union[str, tuple[str, Union[str, bytes], str]]] = {}
                for part in pld.parts:
                    if isinstance(part, payload.MultipartTextField):
                        fields[part.name] = part.value
                    elif isinstance(part, payload.MultipartFileField):
                        fields[part.name] = (part.filename, part.file.read(), part.content_type)

                response = self.http.request_encode_body(r.verb, url, headers=headers, fields=fields, timeout=tmo)

            else:
                raise Exception('unsupported payload type')

        except urllib3_exceptions.ReadTimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        return UResponse(
            status=response.status,
            headers=response.headers,
            data=response.data,
            underlying_object=response,
        )

    def close(self) -> None:
        self.http.clear()

def new_session() -> Session:
    http = urllib3.PoolManager()
    return Session(http)

name: str = urllib3.__name__
version: str = urllib3.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_session=new_session,
)




from __future__ import annotations
from typing import Mapping, Any

from dataclasses import dataclass

from ..artifact import IArtifact

@dataclass(repr=False, eq=False)
class WidgetImageUploadLease(IArtifact):
    d: Mapping[str, Any]
    endpoint: str
    fields: Mapping[str, str]
    s3_object_key: str
    location: str




from __future__ import annotations
from typing import Sequence, Optional

from ..listing.listing_paginator import ListingPaginator
from ....model_loaders.comment_SYNC import load_extra_submission_fields_comment
from ....model_loaders.submission_SYNC import load_submission
from ....exceptions import UnexpectedResultException

class SubmissionAndExtraSubmissionFieldsCommentListingPaginator(ListingPaginator[object]):
    def fetch(self) -> Sequence[object]:
        data = self._fetch_data()
        l = []
        for child in data['children']:
            kind = child['kind']
            data = child['data']
            obj: Optional[object] = None
            if kind == 't1':
                obj = load_extra_submission_fields_comment(data, self.client)
            elif kind == 't3':
                obj = load_submission(data, self.client)
            if obj is None:
                raise UnexpectedResultException(f'unexpected kind {kind!r}')
            l.append(obj)
        return l




from __future__ import annotations
from typing import Mapping, Any

from ..models.wiki_base import BaseWikiPageRevisionAuthorUser, BaseWikiPage, BaseWikiPageRevision, BaseWikiPageSettings


def load_base_wiki_page_revision_author_user(d: Mapping[str, Any]) -> BaseWikiPageRevisionAuthorUser:
    return BaseWikiPageRevisionAuthorUser(d)


def load_base_wiki_page(d: Mapping[str, Any]) -> BaseWikiPage:
    return BaseWikiPage(
        d=d,
        body=d['content_md'],
        body_html=d['content_html'],
        can_revise=d['may_revise'],
        revision_uuid=d['revision_id'],
        revision_timestamp=d['revision_date'],
        revision_author=load_base_wiki_page_revision_author_user(d['revision_by']['data']),
        revision_message=d['reason'] or '',
    )

def load_base_wiki_page_revision(d: Mapping[str, Any]) -> BaseWikiPageRevision:
    return BaseWikiPageRevision(
        d=d,
        uuid=d['id'],
        timestamp=d['timestamp'],
        author=load_base_wiki_page_revision_author_user(d['author']['data']),
        message=d['reason'] or '',
        hidden=d['revision_hidden'],
    )


def load_base_wiki_page_settings(d: Mapping[str, Any]) -> BaseWikiPageSettings:
    return BaseWikiPageSettings(
        permlevel=d['permlevel'],
        # Type ignore due to https://github.com/python/mypy/issues/10986
        editors=[load_base_wiki_page_revision_author_user(m['data']) for m in d['editors']],  # type: ignore[misc]
        unlisted=not d['listed'],
    )




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .....client_SYNC import Client
    from .....models.moderation_action_log_entry import ModerationActionLogEntry
    from ..stream import IStandardStreamEventSubject

from ..stream import Stream


def make_action_log_stream(client: Client, sr: str) -> IStandardStreamEventSubject[ModerationActionLogEntry]:
    it = client.p.moderation.pull_actions(sr)
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.uuid)




from __future__ import annotations
from typing import Iterable, Iterator, Sequence, TypeVar, Callable

from .stubborn_caller_iterator import StubbornCallerIterator
from .unfaltering_chaining_iterator import UnfalteringChainingIterator

T = TypeVar('T')

class CallChunkChainingIterator(UnfalteringChainingIterator[T]):
    """Evaluate call chunks and chain them together."""

    def __init__(self, chunks: Iterable[Callable[[], Sequence[T]]]) -> None:
        self.__chunk_iter: Iterator[Callable[[], Sequence[T]]] = iter(chunks)
        self.__caller_iter: Iterator[Sequence[T]] = StubbornCallerIterator(self.__chunk_iter)
        super().__init__(self.__caller_iter)

    def get_chunk_iter(self) -> Iterator[Callable[[], Sequence[T]]]:
        return self.__chunk_iter

    def get_caller_iter(self) -> Iterator[Sequence[T]]:
        return self.__caller_iter




from __future__ import annotations
from typing import Mapping, Any

from datetime import datetime, timezone

class BaseLiveThread:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.idt: str = d['id']
        self.title: str = d['title']
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.description: str = d['description']
        self.description_html: str = d['description_html']
        self.resources: str = d['resources']
        self.resources_html: str = d['resources_html']
        self.websocket_url: str = d['websocket_url']
        self.complete: bool = d['state'] == 'complete'
        self.nsfw: bool = d['nsfw']
        self.viewer_count: int = d['viewer_count']

class BaseLiveUpdate:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.uuid: str = d['id']
        self.author_name: str = d['author']
        self.body: str = d['body']
        self.body_html: str = d['body_html']
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.stricken: bool = d['stricken']




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .image_size_named_tuple import ImageSize

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class ImageWidgetItem:
    url: str
    size: ImageSize
    link: str




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import ExtraSubmissionFieldsComment

from .pull_ASYNC import Pull

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.listing.comment_listing_async_paginator import ExtraSubmissionFieldsCommentListingAsyncPaginator

class FrontProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.pull: Pull = Pull(client)

    def pull_new_comments(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ExtraSubmissionFieldsCommentListingAsyncPaginator, ExtraSubmissionFieldsComment]:
        p = ExtraSubmissionFieldsCommentListingAsyncPaginator(self._client, '/comments')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import Iterable, Iterator, AsyncIterator, Sequence, Awaitable, TypeVar, Callable

from .stubborn_caller_async_iterator import StubbornCallerAsyncIterator
from .unfaltering_chaining_async_iterator import UnfalteringChainingAsyncIterator

T = TypeVar('T')

class CallChunkChainingAsyncIterator(UnfalteringChainingAsyncIterator[T]):
    """Evaluate call chunks and chain them together."""

    def __init__(self, chunks: Iterable[Callable[[], Awaitable[Sequence[T]]]]) -> None:
        self.__chunk_iter: Iterator[Callable[[], Awaitable[Sequence[T]]]] = iter(chunks)
        self.__caller_iter: AsyncIterator[Sequence[T]] = StubbornCallerAsyncIterator(self.__chunk_iter)
        super().__init__(self.__caller_iter)

    def get_chunk_iter(self) -> Iterator[Callable[[], Awaitable[Sequence[T]]]]:
        return self.__chunk_iter

    def get_caller_iter(self) -> AsyncIterator[Sequence[T]]:
        return self.__caller_iter




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.user_relationship_item import (
        UserRelationshipItem,
        BannedUserRelationshipItem,
    )
    from ...models.moderator_list import ModeratorListItem

from ...model_loaders.moderator_list import load_moderator_list_item
from ...model_loaders.user_relationship_item import load_user_relationship_item, load_banned_user_relation_item

from .legacy_pull_users_SYNC import LegacyPullUsers

class Legacy:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pull_users: LegacyPullUsers = LegacyPullUsers(client)

    def list_moderators(self, sr: str) -> Sequence[ModeratorListItem]:
        root = self._client.request('GET', f'/r/{sr}/about/moderators')
        return [load_moderator_list_item(d) for d in root['data']['children']]

    def get_approved_contributor(self, sr: str, user: str) -> Optional[UserRelationshipItem]:
        root = self._client.request('GET', f'/r/{sr}/about/contributors', params={'user': user})
        children = root['data']['children']
        return load_user_relationship_item(children[0]) if children else None

    def get_approved_wiki_contributor(self, sr: str, user: str) -> Optional[UserRelationshipItem]:
        root = self._client.request('GET', f'/r/{sr}/about/wikicontributors', params={'user': user})
        children = root['data']['children']
        return load_user_relationship_item(children[0]) if children else None

    def get_banned_user(self, sr: str, user: str) -> Optional[BannedUserRelationshipItem]:
        root = self._client.request('GET', f'/r/{sr}/about/banned', params={'user': user})
        children = root['data']['children']
        return load_banned_user_relation_item(children[0]) if children else None

    def get_muted_user(self, sr: str, user: str) -> Optional[UserRelationshipItem]:
        root = self._client.request('GET', f'/r/{sr}/about/muted', params={'user': user})
        children = root['data']['children']
        return load_user_relationship_item(children[0]) if children else None

    def get_wiki_banned_user(self, sr: str, user: str) -> Optional[BannedUserRelationshipItem]:
        root = self._client.request('GET', f'/r/{sr}/about/wikibanned', params={'user': user})
        children = root['data']['children']
        return load_banned_user_relation_item(children[0]) if children else None




from . import (SYNC, ASYNC)  # noqa: F401
from . import (reddit_internal_api_token_obtainment_client_SYNC, reddit_internal_api_token_obtainment_client_ASYNC)  # noqa: F401
from . import const  # noqa: F401
from . import token  # noqa: F401
from .token import Token as Token  # noqa: F401
from . import exceptions  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping, Optional, Protocol
if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from ..session_base_ASYNC import SessionBase

from importlib.util import find_spec
from dataclasses import dataclass

from ...util.imports import load_module_from_spec

class NewSessionFunctionProtocol(Protocol):
    def __call__(self) -> SessionBase: ...

@dataclass
class TransportInfo:
    adaptor_module_name: str
    name: str
    version: str
    new_session: NewSessionFunctionProtocol

def load_spec(name: str, package: Optional[str] = None) -> ModuleSpec:
    module_spec = find_spec(name, package)
    if module_spec is None:
        raise RuntimeError(f'module spec not found: {name} ({package})')
    return module_spec

def load_transport() -> TransportInfo:
    if not transport_registry:
        for module_spec in transport_module_spec_list:
            try:
                load_module_from_spec(module_spec)
            except ImportError:
                continue
            break
        else:
            raise ModuleNotFoundError('An HTTP transport library needs to be installed.')

    return next(iter(transport_registry.values()))

def new_session() -> SessionBase:
    new_session = load_transport().new_session
    return new_session()

def register(
    adaptor_module_name: str,
    new_session: NewSessionFunctionProtocol,
    name: str,
    version: str,
) -> None:
    info = TransportInfo(
        adaptor_module_name=adaptor_module_name,
        name=name,
        version=version,
        new_session=new_session,
    )
    transport_registry[adaptor_module_name] = info

transport_module_spec_list = [
    load_spec('.carriers.httpx_async', __package__),
    load_spec('.carriers.aiohttp', __package__),
]
transport_registry: MutableMapping[str, TransportInfo] = {}




from __future__ import annotations
from typing import TYPE_CHECKING, Any
if TYPE_CHECKING:
    from .token import Token

from ...exceptions import ArgExcMixin

class ArgExc(ArgExcMixin):
    pass


class UnknownTokenType(ArgExc):
    def __init__(self, arg: object = None, *, token: Token):
        super().__init__(arg)
        self.token: Token = token

class RedditInternalAPITokenServerResponseError(ArgExc):
    def __init__(self, arg: object = None, *,
            codename: str = '',
            explanation: str = '') -> None:
        super().__init__(arg)
        self.codename: str = codename
        self.explanation: str = explanation

    def get_default_message(self) -> str:
        co = self.codename
        xp = self.explanation
        if co:
            if xp:
                return f'{co}: {xp}'
            return co
        return ''

def raise_for_token_server_response_error(json_dict: Any) -> None:
    if json_dict.get('success', True):
        return

    error = json_dict['error']
    raise RedditInternalAPITokenServerResponseError(
        codename=error.get('reason') or '',
        explanation=error.get('explanation') or '',
    )




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Any
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ...paginator import MoreAvailablePaginator, CursorPaginator
from ....models.moderation_note import ModerationNote
from ....model_loaders.moderation_note import load_moderation_note


class ModerationNotePaginator(MoreAvailablePaginator[ModerationNote], CursorPaginator[ModerationNote]):
    def __init__(self,
        client: Client,
        uri: str = '/api/mod/notes',
        *,
        limit: Optional[int] = 100,
        cursor: str = '',
        subreddit: str,
        user: str,
        type: Optional[str] = None,
    ):
        super().__init__(limit=limit)
        self.client: Client = client
        self.uri: str = uri
        self._cursor: str = cursor
        self._has_next_page: bool = True
        self._subreddit: str = subreddit
        self._user: str = user
        self._type: Optional[str] = type

    def get_cursor(self) -> str:
        return self._cursor

    def set_cursor(self, value: str) -> None:
        self._cursor = value

    def more_available(self) -> bool:
        return self._has_next_page

    def set_more_available_flag(self, value: bool) -> None:
        self._has_next_page = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        if self.limit is not None:
            yield ('count', str(self.limit))

        if self._cursor:
            yield ('before', self._cursor)

        yield ('subreddit', self._subreddit)
        yield ('user', self._user)
        if self._type:
            yield ('type', self._type)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        root = self.client.request('GET', self.uri, params=params)
        self._cursor = root['end_cursor']
        self._has_next_page = root['has_next_page']
        return root

    def fetch(self) -> Sequence[ModerationNote]:
        root = self._fetch_data()
        mod_notes = root['mod_notes']
        return [load_moderation_note(d) for d in mod_notes]




from __future__ import annotations
from typing import Any, Mapping

from ..models.subreddit_rules import SubredditRules

def load_subreddit_rules(d: Mapping[str, Any]) -> SubredditRules:
    return SubredditRules(d)




from __future__ import annotations
from typing import Sequence, TypeVar, overload, Iterator, Union, Generic

from .submission_draft import Draft
from .subreddit_base import BaseSubreddit

class BaseDraftList(Sequence[Draft]):
    def __init__(self, drafts: Sequence[Draft]):
        self.drafts: Sequence[Draft] = drafts

    def __len__(self) -> int:
        return len(self.drafts)

    def __contains__(self, item: object) -> bool:
        return item in self.drafts

    def __iter__(self) -> Iterator[Draft]:
        return iter(self.drafts)

    @overload
    def __getitem__(self, index: int) -> Draft: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Draft]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Draft, Sequence[Draft]]:
        return self.drafts[index]

TSubreddit = TypeVar('TSubreddit', bound=BaseSubreddit)

class GBaseDraftList(BaseDraftList, Generic[TSubreddit]):
    def __init__(self,
            drafts: Sequence[Draft],
            subreddits: Sequence[TSubreddit]):
        super().__init__(drafts)
        self.subreddits: Sequence[TSubreddit] = subreddits




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_async_iterator import CallChunkCallingAsyncIterator
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...models.modmail_ASYNC import (
    Conversation,
    Message,
    ConversationAggregate,
    UserDossierConversationAggregate,
    OptionalUserDossierConversationAggregate,
)
from ...model_loaders.modmail_ASYNC import (
    load_conversation,
    load_message,
    load_conversation_aggregate,
    load_user_dossier_conversation_aggregate,
    load_optional_user_dossier_conversation_aggregate,
)
from .conversation_pulls_ASYNC import Pulls

class ConversationProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pulls: Pulls = Pulls(client)

    async def fetch(self, idn: int, *, mark_read: bool = False) -> OptionalUserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('GET', f'/api/mod/conversations/{convo_id36}')
        return load_optional_user_dossier_conversation_aggregate(
            root['conversation'],
            root['messages'],
            root['modActions'],
            root['user'] or None,
            client=self._client,
        )

    async def create(self, sr: str, to: str, subject: str, body: str, *, hidden: bool = False) -> tuple[Conversation, Message]:
        data = {'srName': sr, 'to': to, 'subject': subject, 'body': body, 'isAuthorHidden': '01'[hidden]}
        root = await self._client.request('POST', '/api/mod/conversations', data=data)
        conversation_data = root['conversation']
        conversation = load_conversation(conversation_data, self._client)
        message_id36 = conversation_data['objIds'][0]['id']
        message = load_message(root['messages'][message_id36], self._client)
        return (conversation, message)

    async def reply(self, idn: int, body: str, *, hidden: bool = False, internal: bool = False) -> ConversationAggregate:
        data = {'body': body, 'isAuthorHidden': '01'[hidden], 'isInternal': '01'[internal]}
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}', data=data)
        conversation_data = root['conversation']
        messages_mapping_data = root['messages']
        mod_actions_mapping_data = conversation_data['modActions']
        return load_conversation_aggregate(
            conversation_data,
            messages_mapping_data,
            mod_actions_mapping_data,
            client=self._client,
        )

    async def mark_read(self, idn: int) -> None:
        data = {'conversationIds': to_base36(idn)}
        await self._client.request('POST', '/api/mod/conversations/read', data=data)

    def bulk_mark_read(self, ids: Iterable[int]) -> CallChunkCallingAsyncIterator[None]:
        async def mass_mark_read(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            ids_str = ','.join(id36s)
            await self._client.request('POST', '/api/mod/conversations/read', data={'conversationIds': ids_str})

        return CallChunkCallingAsyncIterator(AsyncCallChunk(mass_mark_read, chunk) for chunk in chunked(ids, 100))

    async def mark_unread(self, idn: int) -> None:
        data = {'conversationIds': to_base36(idn)}
        await self._client.request('POST', '/api/mod/conversations/unread', data=data)

    def bulk_mark_unread(self, ids: Iterable[int]) -> CallChunkCallingAsyncIterator[None]:
        async def mass_mark_unread(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            ids_str = ','.join(id36s)
            await self._client.request('POST', '/api/mod/conversations/unread', data={'conversationIds': ids_str})

        return CallChunkCallingAsyncIterator(AsyncCallChunk(mass_mark_unread, chunk) for chunk in chunked(ids, 100))

    def bulk_mark_all_read(self, mailbox: str, subrs: Iterable[str]) -> CallChunkChainingAsyncIterator[int]:
        async def mass_mark_all_read(subrs: Sequence[str]) -> Sequence[int]:
            subrs_str = ','.join(subrs)
            data = {'state': mailbox, 'entity': subrs_str}
            root = await self._client.request('POST', '/api/mod/conversations/bulk/read', data=data)
            return [int(id36, 36) for id36 in root['conversation_ids']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_mark_all_read, chunk) for chunk in chunked(subrs, 100))

    async def highlight(self, idn: int) -> ConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/highlight')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    async def unhighlight(self, idn: int) -> ConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('DELETE', f'/api/mod/conversations/{convo_id36}/highlight')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    async def archive(self, idn: int) -> ConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/archive')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    async def unarchive(self, idn: int) -> ConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unarchive')
        return load_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            client=self._client,
        )

    async def approve_user(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/approve')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def disapprove_user(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/disapprove')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def mute_user_3d(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '72'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def mute_user_7d(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '168'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def mute_user_28d(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/mute', data={'num_hours': '672'})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def unmute_user(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unmute')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def unban_user(self, idn: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/unban')
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )

    async def shorten_user_ban(self, idn: int, duration: int) -> UserDossierConversationAggregate:
        convo_id36 = to_base36(idn)
        root = await self._client.request('POST', f'/api/mod/conversations/{convo_id36}/temp_ban',
                data={'duration': str(duration)})
        return load_user_dossier_conversation_aggregate(
            root['conversations'],
            root['messages'],
            root['modActions'],
            root['user'],
            client=self._client,
        )




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...http.requestor_ASYNC import Requestor
    from ...http.request import Request
    from ...http.response import Response

from ...util.imports import lazy_import
if TYPE_CHECKING:
    import asyncio
else:
    lazy_import % 'asyncio'

from ...http.requestor_augmenter_ASYNC import RequestorAugmenter
from ...util.token_bucket import TokenBucket

class RateLimited(RequestorAugmenter):
    def __init__(self, requestor: Requestor) -> None:
        super().__init__(requestor)
        self._tb = TokenBucket(3, 1/2)

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        await asyncio.sleep(self._tb.get_cooldown(1))
        self._tb.consume(1)
        return await self.requestor.send(request, timeout=timeout)




from __future__ import annotations

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class CalendarWidgetConfiguration:
    num_events: int
    show_title: bool
    show_description: bool
    show_location: bool
    show_date: bool
    show_time: bool




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...util.base_conversion import to_base36

class Reorder:
    def __init__(self, client: Client):
        self._client = client

    def __call__(self, uuid: str, submission_ids: Sequence[int]) -> None:
        submission_id36s = [to_base36(i) for i in submission_ids]
        return self.by_id36(uuid, submission_id36s)

    def by_id36(self, uuid: str, submission_id36s: Sequence[str]) -> None:
        submission_full_id36s = ['t3_' + s for s in submission_id36s]
        link_ids_str = ','.join(submission_full_id36s)
        params = {'collection_id': uuid, 'link_ids': link_ids_str}
        self._client.request('POST', '/api/v1/collections/reorder_collection', params=params)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..http.requestor_SYNC import Requestor
    from ..http.request import Request
    from ..http.response import Response

import time

from ..http.requestor_augmenter_SYNC import RequestorAugmenter
from ..util.token_bucket import TokenBucket

class RateLimited(RequestorAugmenter):
    def __init__(self, requestor: Requestor) -> None:
        super().__init__(requestor)
        self.reset: int = 0
        self.remaining: int = 0
        self.used: int = 0
        self._delta: float = 0.
        self._timestamp: float = time.monotonic()
        self._burst_control = TokenBucket(5, 1/2)

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        h = self._burst_control.hard_consume(1)
        s = 0.
        if self.remaining < 2:
            s = self.reset
        elif h and (w := self.reset / self.remaining) < 2:
            # Use 2 because at worst the user is allocated a 2/s
            # rate limit when a client credentials grant is used.
            s = w

        time.sleep(s)

        response = self.requestor.send(request, timeout=timeout)

        now = time.monotonic()
        self._delta = now - self._timestamp
        self._timestamp = now

        headers = response.headers
        if 'x-ratelimit-reset' in headers:
            self.reset = int(headers['x-ratelimit-reset'])
            self.remaining = int(float(headers['x-ratelimit-remaining']))
            self.used = int(headers['x-ratelimit-used'])
        else:
            if self.reset > 0:
                self.reset = max(self.reset - int(self._delta), 0)
                self.remaining -= 1
                self.used += 1
            else:
                self.reset = 600
                self.remaining = 300
                self.used = 0

        return response




from __future__ import annotations
from typing import Sequence

from ..listing.listing_paginator import ListingPaginator
from ....model_loaders.subreddit_SYNC import load_subreddit
from ....models.subreddit_SYNC import Subreddit

class SubredditListingPaginator(ListingPaginator[Subreddit]):
    def fetch(self) -> Sequence[Subreddit]:
        data = self._fetch_data()
        return [load_subreddit(d['data'], self.client) for d in data['children']]



#!/usr/bin/env python3
"""
Step through the OAuth2 Authorization Code Flow to obtain OAuth2 tokens.

Visit <https://www.reddit.com/prefs/apps> to create an app for your client.

Your app's redirect URI must *exactly* match `http://localhost:8080`.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from types import FrameType


import argparse
class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
globalz = globals()
cli_desc = globalz.get('?description', __doc__)
parser = argparse.ArgumentParser(description=cli_desc, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
add = parser.add_argument
add('client_id', nargs='?')
add('client_secret', nargs='?')
add('--client-id', metavar='CLIENT_ID', dest='client_id_opt', help=argparse.SUPPRESS)
add('--client-secret', metavar='CLIENT_SECRET', dest='client_secret_opt', help=argparse.SUPPRESS)
add('--scope', default='*', help="an OAuth2 scope string")
add('--redirect-uri', default="http://localhost:8080", help="\N{ZERO WIDTH SPACE}")
if globalz.get('?access_token_only', False):
    add('--access-token-only', default=True, help=argparse.SUPPRESS)
else:
    add('--access-token-only', action='store_true', help="get an access token and not a refresh token")
add('--authorization-code-only', action='store_true', help=argparse.SUPPRESS)
add('--no-web-browser', action='store_true', help="don't launch web browser")
add('--web-browser-name', help="the web browser to open")
args = parser.parse_args()


import sys
import os
import re
import uuid
import socket
import urllib.parse
import webbrowser
import signal
from functools import partial
from pprint import pp

import redditwarp
from redditwarp.http.transport.SYNC import load_transport, new_session
from redditwarp.core.SYNC import RedditTokenObtainmentClient
from redditwarp.util.user_agent_SYNC import get_user_agent_from_session

def get_client_cred_input(v: Optional[str], prompt: str, env: str) -> str:
    if v is None:
        v = input(prompt)
    if v == '.':
        v = os.environ[env]
        print(v)
    return v

if not sys.flags.interactive:
    @partial(signal.signal, signal.SIGINT)
    def handle_sigint(sig: int, frame: Optional[FrameType]) -> None:
        print('KeyboardInterrupt', file=sys.stderr)
        sys.exit(130)

load_transport()

client_id = get_client_cred_input(
        (args.client_id_opt or args.client_id),
        'Client ID: ', 'redditwarp_client_id')
client_secret = get_client_cred_input(
        (args.client_secret_opt or args.client_secret),
        'Client secret: ', 'redditwarp_client_secret')
scope: str = args.scope
redirect_uri: str = args.redirect_uri
access_token_only: bool = args.access_token_only
authorization_code_only: bool = args.authorization_code_only
no_web_browser: bool = args.no_web_browser
web_browser_name: Optional[str] = args.web_browser_name
state = str(uuid.uuid4())

port = urllib.parse.urlsplit(redirect_uri).port
if port is None:
    raise Exception('could not extract a port number from the redirect URI')

browser = webbrowser.get(web_browser_name)

print('''\n        -~=~- Reddit OAuth2 Authorization Code Flow -~=~-\n
* Step 1. Build the authorization URL and direct the user to the authorization server.\n''')

params = {
    'response_type': 'code',
    'client_id': client_id,
    'redirect_uri': redirect_uri,
    'scope': scope,
    'state': state,
    'duration': 'temporary' if access_token_only else 'permanent',
}
url = "%s?%s" % (redditwarp.auth.const.AUTHORIZATION_URL, urllib.parse.urlencode(params))
print(url)

if not no_web_browser:
    browser.open(url)

print('''\n* Step 2. Wait for the authorization server response and extract the authorization code.\n
Abort with ^C if the authorization request was rejected.\n''')

match = None

with socket.socket() as server:
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(('localhost', port))
    server.listen()

    while not match:
        client, addr = server.accept()
        print('{0}:{1}'.format(*addr))

        with client:
            data = client.recv(8192)
            client.sendall(b"HTTP/1.1 200 OK\r\n\r\n" + data)

        text = data.decode()
        print(f"```\n{text}\n```\n")
        match = re.match(r"^GET [^?]*\?([^ ]*) HTTP", text)

assert match

response_params = dict(urllib.parse.parse_qsl(match[1]))

received_state = response_params['state']
if received_state != state:
    raise Exception(f'sent state ({state}) did not match received state ({received_state})')

code = response_params.get('code', '')
if not code:
    raise Exception('authorization was declined by the user')
print(f'Authorization code: {code}')

if authorization_code_only:
    sys.exit(0)

print('\n* Step 3. Exchange the authorization code for an access/refresh token.\n')

grant = redditwarp.auth.grants.AuthorizationCodeGrant(code, redirect_uri)
print('Authorization grant:')
pp(dict(grant))
print()

session = new_session()
user_agent = get_user_agent_from_session(session) + " redditwarp.cli.refresh_token"
token_client = RedditTokenObtainmentClient(
    session,
    redditwarp.auth.const.TOKEN_OBTAINMENT_URL,
    (client_id, client_secret),
    grant,
    headers={'User-Agent': user_agent},
)

print('Obtaining token(s) from token server...\n')

token = token_client.fetch_token()

print(f'''\
     Access token: {token.access_token}
    Refresh token: {token.refresh_token}''')




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping
if TYPE_CHECKING:
    from ..request import Request
    from ..response import Response

from ..requestor_ASYNC import Requestor
from ..requestor_augmenter_ASYNC import RequestorAugmenter

class DestinationNotEstablishedException(Exception):
    pass

class RequestDistributor(Requestor):
    def __init__(self) -> None:
        super().__init__()
        self.directions: MutableMapping[Request, Requestor] = {}

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        try:
            dest = self.directions[request]
        except KeyError:
            raise DestinationNotEstablishedException from None
        return await dest.send(request)

    def get_director(self, sender: Requestor, receiver: Requestor) -> RequestDirector:
        return RequestDirector(sender, self, receiver)

class RequestDirector(RequestorAugmenter):
    def __init__(self, requestor: Requestor, target: RequestDistributor, receiver: Requestor):
        super().__init__(requestor)
        self.target: RequestDistributor = target
        self.receiver: Requestor = receiver

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        directions = self.target.directions
        directions[request] = self.receiver
        try:
            resp = await self.requestor.send(request)
        finally:
            del directions[request]
        return resp




from __future__ import annotations
from typing import Mapping, Any

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class PostFlairWidgetItem:
    d: Mapping[str, Any]
    uuid: str
    type: str
    text: str
    bg_color: str
    fg_light_or_dark: str




from __future__ import annotations
from typing import Any, Mapping

from ..models.subreddit_user_item import (
    ModeratorUserItem,
    ContributorUserItem,
    BannedUserItem,
    MutedUserItem,
)

def load_moderator_user_item(d: Mapping[str, Any]) -> ModeratorUserItem:
    return ModeratorUserItem(d)

def load_contributor_user_item(d: Mapping[str, Any]) -> ContributorUserItem:
    return ContributorUserItem(d)

def load_banned_user_item(d: Mapping[str, Any]) -> BannedUserItem:
    return BannedUserItem(d)

def load_muted_user_item(d: Mapping[str, Any]) -> MutedUserItem:
    return MutedUserItem(d)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable
if TYPE_CHECKING:
    from ..auth.token_obtainment_client_SYNC import TokenObtainmentClient
    from ..http.requestor_SYNC import Requestor
    from ..http.request import Request
    from ..http.response import Response

import time

from ..auth.token import Token
from ..http.requestor_augmenter_SYNC import RequestorAugmenter
from ..auth.exceptions import (
    UnknownTokenType,
    extract_www_authenticate_auth_params,
    raise_for_resource_server_response_error,
)

class Authorizer:
    def __init__(self,
        token_client: Optional[TokenObtainmentClient] = None,
        token: Optional[Token] = None,
    ):
        self.token_client: Optional[TokenObtainmentClient] = token_client
        self.token: Optional[Token] = token
        self.renewal_time: Optional[int] = None
        self.renewal_skew: int = 30
        self.expires_in_fallback: Optional[int] = None
        self.time_func: Callable[[], float] = time.monotonic
        self.authorization_header_name: str = 'Authorization'

    def time(self) -> float:
        return self.time_func()

    def renew_token(self) -> None:
        if self.token_client is None:
            raise RuntimeError('no token client')

        self.token = tk = self.token_client.fetch_token()

        if tk.token_type.lower() != 'bearer':
            raise UnknownTokenType(token=tk)

        expires_in: Optional[int] = tk.expires_in
        if expires_in is None:
            expires_in = self.expires_in_fallback
        if expires_in is None:
            self.renewal_time = None
        else:
            self.renewal_time = int(self.time()) + expires_in - self.renewal_skew

        if tk.refresh_token:
            grant1 = self.token_client.grant
            if (
                grant1.get('grant_type', '') == 'refresh_token'
                and grant1.get('refresh_token', '') != tk.refresh_token
            ):
                self.token_client.grant = {**grant1, 'refresh_token': tk.refresh_token}

    def should_renew_token(self) -> bool:
        if self.token is None:
            return True
        if self.renewal_time is None:
            return False
        return self.time() >= self.renewal_time

    def can_renew_token(self) -> bool:
        return self.token_client is not None

    def get_token(self) -> Token:
        tk = self.token
        if tk is None:
            raise RuntimeError('token not set')
        return tk

    def set_token(self, token: Token) -> None:
        self.token = token

    def set_access_token(self, access_token: str) -> None:
        self.set_token(Token(access_token))

    def prepare_request(self, request: Request, token: Optional[Token] = None) -> None:
        tk = self.get_token() if token is None else token
        request.headers[self.authorization_header_name] = f'{tk.token_type} {tk.access_token}'


class Authorized(RequestorAugmenter):
    """Used to perform requests to endpoints that require authorization."""

    def __init__(self, requestor: Requestor, authorizer: Authorizer) -> None:
        super().__init__(requestor)
        self.authorizer: Authorizer = authorizer

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        authorizer = self.authorizer
        if authorizer.should_renew_token():
            authorizer.renew_token()

        authorizer.prepare_request(request)
        resp = self.requestor.send(request, timeout=timeout)
        auth_params = extract_www_authenticate_auth_params(resp)

        invalid_token = auth_params.get('error', '') == 'invalid_token'
        if invalid_token and authorizer.can_renew_token():
            authorizer.renew_token()
            authorizer.prepare_request(request)
            resp = self.requestor.send(request, timeout=timeout)
            auth_params = extract_www_authenticate_auth_params(resp)

        raise_for_resource_server_response_error(auth_params)

        return resp
